unit GlowControls;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls,
  Graphics, ExtCtrls, Buttons, Math, Forms;

type
  TFadeSpeed = (fsSlow, fsMedium, fsFast, fsVeryFast);
  TGradientType = (gtHorizontal, gtVertical, gtFromTopLeft, gtFromTopRight,
                   gtRadial, gtDoubleHorz, gtDoubleVert);
  TRGBArray = array[0..0] of TRGBTriple;
  PRGBArray = ^TRGBArray;
  TDrawStates = set of (dsMouseOver, dsClicking);
  TBorderLines = set of (blTop, blBottom, blLeft, blRight);
  TShCaptionPos = (cpGroupBox, cpTopLeft, cpTopCenter, cpTopRight,
                 cpMiddleLeft, cpMiddleCenter, cpMiddleRight,
                 cpBottomLeft, cpBottomCenter, cpBottomRight);
  TBtnStyle   = (bsNormal,bsRect,bsRT,bsRTRB,bsRB,bsLT,bsLTLB,bsLB);
  TSlideMode  = (smNone,smFull,smLeft,smRight,smTop,smBottom);
  TSlideStyle = (ssNormal,ssCombine);
  TGradientStyle = (gsNone,gsLeftVertical,gsRightVertical,gsTopHorizontal,gsBottomHorizontal);
const
    AFadeSpeed : array[TFadeSpeed] of integer = (40, 20, 12, 8);
    ClickBlendInterval = 25;
    OverBlendInterval = 10;
type

  TShCustomColors = class(TPersistent)
  private
    FDefaultFrom : TColor;
    FDefaultTo : TColor;
    FOverFrom : TColor;
    FOverTo : TColor;
    procedure SetDefaultFrom(const Value: TColor);
    procedure SetDefaultTo(const Value: TColor);
    procedure SetOverFrom(const Value: TColor);
    procedure SetOverTo(const Value: TColor);
  protected
    FOwner: TObject;
    procedure DoUpdateGradients; virtual; abstract;
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create(AOwner: TComponent); virtual;
  published
    property DefaultFrom : TColor read FDefaultFrom write SetDefaultFrom;
    property DefaultTo : TColor read FDefaultTo write SetDefaultTo;
    property OverFrom : TColor read FOverFrom write SetOverFrom;
    property OverTo : TColor read FOverTo write SetOverTo;
  end;

  TShControlColors = class(TShCustomColors)
  private
    FClickedFrom : TColor;
    FClickedTo : TColor;
    FHotTrack : TColor;
    FBorderColor : TColor;
    FTextShadow : TColor;
    procedure SetClickedFrom(const Value: TColor);
    procedure SetClickedTo(const Value: TColor);
    procedure SetHotTrack(const Value: TColor);
    procedure SetBorderColor(const Value: TColor);
    procedure SetTextShadow(const Value: TColor);
  protected
    FOwner: TObject;
    procedure DoUpdateGradients; override;
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property ClickedFrom : TColor read FClickedFrom write SetClickedFrom;
    property ClickedTo : TColor read FClickedTo write SetClickedTo;
    property HotTrack : TColor read FHotTrack write SetHotTrack;
    property BorderColor : TColor read FBorderColor write SetBorderColor;
    property TextShadow : TColor read FTextShadow write SetTextShadow;
  end;

  TShCustomControl = class(TCustomControl)
  private
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FTextShadow: boolean;
    FShowCaption: boolean;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMFocusChanged(var Message: TMessage); message CM_FOCUSCHANGED;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMParentColorChanged(var Message: TMessage); message CM_PARENTCOLORCHANGED;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure WMMouseMove(var Message: TWMMouse); message WM_MOUSEMOVE;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure SetTextShadow(const Value: boolean);
    procedure SetShowCaption(const Value: boolean);
  protected
    FOverBlendPercent: integer;
    FClickedBlendPercent: integer;
    FDrawStates: TDrawStates;
    FFadeSpeed: TFadeSpeed;
    FOnFade : TNotifyEvent;
    procedure UpdateGradients; virtual; abstract;
    procedure EventEnabledChanged; virtual;
    procedure EventFocusChanged; virtual;
    procedure EventMouseDown; virtual;
    procedure EventMouseUp; virtual;
    procedure EventMouseEnter; virtual;
    procedure EventMouseLeave; virtual;
    procedure EventMouseMove; virtual;
    procedure EventParentColorChanged; virtual;
    procedure EventResized; virtual;
    procedure EventTextChanged; virtual;
    procedure DoBlendLess(Sender: TObject); virtual;
    procedure DoBlendMore(Sender: TObject); virtual;
    procedure UpdateTimer;
    procedure WndProc(var Msg: TMessage); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure MouseDown(Button:TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button:TMouseButton; Shift:TShiftState; X, Y: Integer); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  published
    property Anchors;
    property Action;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
    property OnEndDock;
    property OnStartDock;
    property Align;
    property Color;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property TabOrder;
    property TabStop default True;
    property TextShadow: boolean read FTextShadow write SetTextShadow;
    property ShowCaption: boolean read FShowCaption write SetShowCaption;
  end;

  procedure DrawGradient(Canvas: TCanvas; FromColor, ToColor: TColor; Steps: Integer; R: TRect; Direction: Boolean);
  procedure DoBitmapGradient(Bitmap: TBitmap; CurrRect: TRect; Color1, Color2: TColor; GradientType : TGradientType; Region: HRGN);
  procedure DoDrawText(Canvas: TCanvas; AText: string; AFont : TFont; Enabled : boolean; ShowAccelChar : boolean; ARect: TRect; Flags: integer; TextShadow: boolean; ShadowColor : TColor);
  procedure BlendBitmaps(AFinal, ABitmap1, ABitmap2 : TBitmap; APercent : integer; ARect: TRect; Offset : integer);
  procedure SetTextFlags(Alignment: TAlignment; WordWrap: boolean; var Flags: integer);
  procedure ColorizeBitmap(ABitmap: TBitmap; Color : TColor);
  procedure AntialiasBitmap(ABitmap1, ABitmap2: TBitmap; Sampling: integer);
  procedure AntialiasShapeBitmap(ABitmap1, ABitmap2: TBitmap; Sampling: integer; CornerWidth: integer; InnerColor: TColor);
  procedure RbAlphaBlend(ABitmap1, ABitmap2: TBitmap; ARect: TRect; percent : integer);

type
  TShButton = class(TShCustomControl)
  private
    FColors : TShControlColors;
    FGrBitmap: TBitmap;
    FGrMouseOver: TBitmap;
    FGrClicked: TBitmap;
    FGlyph : TBitmap;
    FLayout : TButtonLayout;
    FModalResult: TModalResult;
    FSpacing: integer;
    FShowFocusRect : boolean;
    FHotTrack : boolean;
    FDefault: boolean;
    FCancel : boolean;
    FGradientBorder: boolean;
    FDown: boolean;
    FAllowAllUp: boolean;
    FGroupIndex: integer;
    FGradientType: TGradientType;
    FFadeSpeed: TFadeSpeed;
    FBtnStyle: TBtnStyle;
    procedure CMDialogKey(var Message: TCMDialogKey); message CM_DIALOGKEY;
    procedure SetGlyph(const Value: TBitmap);
    procedure SetSpacing(const Value: integer);
    procedure SetLayout(const Value: TButtonLayout);
    procedure SetFadeSpeed(const Value: TFadeSpeed);
    procedure SetFocusRect(const Value: boolean);
    procedure SetHotTrack(const Value: boolean);
    procedure SetDefault(const Value: boolean);
    procedure SetAllowAllUp(const Value: boolean);
    procedure SetDown(const Value: boolean);
    procedure SetGroupIndex(const Value: integer);
    procedure DeActivateSiblings;
    function  CanBeUp: boolean;
    procedure SetGradientType(const Value: TGradientType);
    procedure SetBtnStyle(const Value: TBtnStyle);
  protected
    procedure Paint; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure EventResized; override;
    procedure EventMouseEnter; override;
    procedure EventMouseLeave; override;
    procedure EventMouseDown; override;
    procedure EventMouseUp; override;
    procedure EventFocusChanged; override;
    procedure UpdateGradients; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Click; override;
  published
    property Caption;
    property ModalResult : TModalResult read FModalResult write FModalResult;
    property Glyph: TBitmap read FGlyph write SetGlyph;
    property Spacing: integer read FSpacing write SetSpacing;
    property Layout: TButtonLayout read FLayout write SetLayout;
    property Colors : TShControlColors read FColors write FColors;
    property FadeSpeed : TFadeSpeed read FFadeSpeed write SetFadeSpeed;
    property ShowFocusRect : boolean read FShowFocusRect write SetFocusRect;
    property HotTrack: boolean read FHotTrack write SetHotTrack;
    property Default : boolean read FDefault write SetDefault default false;
    property Cancel : boolean read FCancel write FCancel default false;
    property GradientBorder: boolean read FGradientBorder write FGradientBorder;
    property GroupIndex: integer read FGroupIndex write SetGroupIndex;
    property AllowAllUp: boolean read FAllowAllUp write SetAllowAllUp;
    property Down : boolean read FDown write SetDown;
    property GradientType: TGradientType read FGradientType write SetGradientType;
    property BtnStyle: TBtnStyle read FBtnStyle write SetBtnStyle default bsRect;
  end;

  TShCustomPanel = class(TCustomControl)
      fTimer       : TTimer;
    private
      fSlideMode  : TSlideMode;
      fSlideStyle : TSlideStyle;
      fGlyph      : TBitmap;
      fCollapse   : boolean;
      fGradient   : TGradientStyle;
      procedure   TimerTimer(Sender: TObject);
      procedure   SetGlyph(const Value: TBitmap);
      procedure   SetCollapse(const Value: boolean);
    protected
      fMin        : integer;
      fWidthMax   : integer;
      fHeightMax  : integer;
      fStep       : integer;
      fKoeff      : integer;
      fRegion     : boolean;
      fClosed     : boolean;
      fGlyphOn    : boolean;
      procedure   Paint; override;
      procedure   WMSize(var Message: TWMSize); message WM_SIZE;
      procedure   MouseMove(Shift: TShiftState; X, Y: Integer); override;
      procedure   MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      procedure   SetMinMax(Min,WMax,HMax: integer);
    published
      property    Glyph: TBitmap read fGlyph write SetGlyph;
      property    Collapse: boolean read fCollapse write SetCollapse;
      property    Close: boolean read fClosed;
      property    SlideMode: TSlideMode read fSlideMode write fSlideMode;
      property    SlideStyle: TSlideStyle read fSlideStyle write fSlideStyle;
      property    Gradient: TGradientStyle read fGradient write fGradient;
    end;

  TShPanel = class(TShCustomPanel)
    published
      property    Collapse;
      property    Color;
      property    Glyph;
      property    SlideMode;
      property    SlideStyle;
      property    OnClick;
      property    OnDblClick;
      property    Close;
    end;

procedure Register;

implementation

{$R GlowControls.RES}

procedure Register;
begin
  RegisterComponents('Soft Universe', [TShButton]);
end;

const
  sFileNotExists = 'File not exists!';
  sInvalidOrEmptyFile = 'Empty or invalid file';

procedure DrawGradient(Canvas: TCanvas; FromColor, ToColor: TColor; Steps: Integer; R: TRect; Direction: Boolean);
var
  diffr, startr, endr: Integer;
  diffg, startg, endg: Integer;
  diffb, startb, endb: Integer;
  rstepr, rstepg, rstepb, rstepw: Real;
  i, stepw: Word;

begin
  if Steps = 0 then
    Steps := 1;

  FromColor := ColorToRGB(FromColor);
  ToColor := ColorToRGB(ToColor);

  startr := (FromColor and $0000FF);
  startg := (FromColor and $00FF00) shr 8;
  startb := (FromColor and $FF0000) shr 16;
  endr := (ToColor and $0000FF);
  endg := (ToColor and $00FF00) shr 8;
  endb := (ToColor and $FF0000) shr 16;

  diffr := endr - startr;
  diffg := endg - startg;
  diffb := endb - startb;

  rstepr := diffr / steps;
  rstepg := diffg / steps;
  rstepb := diffb / steps;

  if Direction then
    rstepw := (R.Right - R.Left) / Steps
  else
    rstepw := (R.Bottom - R.Top) / Steps;

  with Canvas do
  begin
    for i := 0 to steps - 1 do
    begin
      endr := startr + Round(rstepr * i);
      endg := startg + Round(rstepg * i);
      endb := startb + Round(rstepb * i);
      stepw := Round(i * rstepw);
      Pen.Color := endr + (endg shl 8) + (endb shl 16);
      Brush.Color := Pen.Color;
      if Direction then
        Rectangle(R.Left + stepw, R.Top, R.Left + stepw + Round(rstepw) + 1, R.Bottom)
      else
        Rectangle(R.Left, R.Top + stepw, R.Right, R.Top + stepw + Round(rstepw) + 1);
    end;
  end;
end;

procedure DoBitmapGradient(Bitmap: TBitmap; CurrRect: TRect; Color1, Color2: TColor; GradientType : TGradientType; Region: HRGN);
var
  y, x, MaxDist, C1, C2 : integer;
  rowDest : PRGBArray;
  Dep : array[0..2] of integer;
  Dif : array[0..2] of Double;
  Coef : Double;
begin

  Coef:=0;
  MaxDist:=1;

  if (Bitmap.Width <= 0) or (Bitmap.Height <= 0) then Exit;

  if Color1 = Color2  then begin
    Bitmap.Canvas.Brush.Color := Color1;
    if Region > 0 then
      FillRgn(Bitmap.Canvas.Handle, Region, Bitmap.Canvas.Brush.Handle)
    else
      Bitmap.Canvas.FillRect(CurrRect);
    Exit;
  end;

  C1 := ColorToRGB(Color1);
  C2 := ColorToRGB(Color2);

  case GradientType of
    gtVertical : MaxDist := CurrRect.Bottom - CurrRect.Top;

    gtHorizontal : MaxDist := CurrRect.Right - CurrRect.Left;

    gtFromTopLeft,
    gtFromTopRight: MaxDist := (Round(Sqrt( Sqr(CurrRect.Right - CurrRect.Left)
                              + Sqr(CurrRect.Bottom - CurrRect.Top)))
                              - CurrRect.Left);
    gtRadial: MaxDist := (Round(Sqrt( Sqr(CurrRect.Right - CurrRect.Left)
                              + Sqr(CurrRect.Bottom - CurrRect.Top)))
                              - CurrRect.Left) div 2;
    gtDoubleHorz: MaxDist :=  (CurrRect.Right - CurrRect.Left) div 2;
    gtDoubleVert: MaxDist :=  (CurrRect.Bottom - CurrRect.Top) div 2;
  end;

  Dep[0] := GetRValue(C1);
  Dep[1] := GetGValue(C1);
  Dep[2] := GetBValue(C1);

  Dif[0] := (GetRValue(C2) - Dep[0]) / MaxDist;
  Dif[1] := (GetGValue(C2) - Dep[1]) / MaxDist;
  Dif[2] := (GetBValue(C2) - Dep[2]) / MaxDist;

  if Dif[0] > 255 then Dif[0] := 255;
  if Dif[1] > 255 then Dif[1] := 255;
  if Dif[2] > 255 then Dif[2] := 255;

  for y:= CurrRect.Top to CurrRect.Bottom -1 do
  begin
    rowDest := Bitmap.ScanLine[y];
    for x := CurrRect.Left to CurrRect.Right - 1 do
    begin
      if (Region > 0) and not PtInRegion(Region, x, y) then Continue;
      case GradientType of
        gtHorizontal : Coef := x - CurrRect.Left;
        gtVertical :   Coef := y - CurrRect.Top;
        gtFromTopLeft :Coef := Sqrt(Sqr(x - CurrRect.Left) + Sqr(y - CurrRect.Top));
        gtFromTopRight:Coef := Sqrt(Sqr(CurrRect.Right - x) + Sqr(y - CurrRect.Top));
        gtRadial: Coef := Sqrt(Sqr(((CurrRect.Right - CurrRect.Left) shr 1) - x) + Sqr(y - ((CurrRect.Bottom - CurrRect.Top) shr 1)));
        gtDoubleHorz: Coef := Abs(x - MaxDist - CurrRect.Left);
        gtDoubleVert: Coef := Abs(y - MaxDist - CurrRect.Top);
      end;
      RowDest[x].rgbtRed := Byte(Round(Dep[0] + (Dif[0] * Coef)));
      RowDest[x].rgbtGreen := Byte(Round(Dep[1] + (Dif[1] * Coef)));
      RowDest[x].rgbtBlue := Byte(Round(Dep[2] + (Dif[2] * Coef)));
    end;
 end;
end;

procedure DoDrawText(Canvas: TCanvas; AText: string; AFont : TFont; Enabled : boolean; ShowAccelChar : boolean; ARect: TRect; Flags: integer; TextShadow: boolean; ShadowColor : TColor);
var
  OldColor : TColor;
begin
  with Canvas do begin
    Canvas.Font.Assign(AFont);
    if Enabled then begin
      if TextShadow then begin
        OldColor := Font.Color;
        Font.Color := ShadowColor;
        OffsetRect(ARect, 1, 1);
        DrawTextEx(Handle, PChar(AText), Length(AText), ARect, Flags, nil);
        Font.Color := OldColor;
        OffsetRect(ARect, -1, -1);
      end;
      DrawTextEx(Handle, PChar(AText), Length(AText), ARect, Flags, nil);

    end else begin
      Font.Color := clBtnHighlight;
      OffsetRect(ARect, 1, 1);
      DrawTextEx(Handle, PChar(AText), Length(AText), ARect, Flags, nil);

      Font.Color := clBtnShadow;
      OffsetRect(ARect, -1, -1);
      DrawTextEx(Handle, PChar(AText), Length(AText), ARect, Flags, nil);
    end;
  end;
end;

procedure BlendBitmaps(AFinal, ABitmap1, ABitmap2 : TBitmap; APercent : integer; ARect: TRect; Offset : integer);
begin
  BitBlt(AFinal.Canvas.Handle, ARect.Left + Offset, ARect.Top + Offset, ARect.Right - ARect.Left - Offset, ARect.Bottom - ARect.Top - Offset,
         ABitmap1.Canvas.Handle, 0, 0, SRCCOPY);
  if APercent <= 0 then Exit;
  if APercent > 100 then APercent := 100;
  RbAlphaBlend(AFinal, ABitmap2, ARect, APercent);
end;

procedure SetTextFlags(Alignment: TAlignment; WordWrap: boolean; var Flags: integer);
begin
  Flags := DT_END_ELLIPSIS or DT_VCENTER or DT_NOPREFIX;

  case Alignment of
    taLeftJustify:
      Flags := Flags or DT_LEFT;
    taCenter:
      Flags := Flags or DT_CENTER;
    taRightJustify:
      Flags := Flags or DT_RIGHT;
  end;

  if WordWrap then
    Flags := Flags or DT_WORDBREAK
  else
    Flags := Flags or DT_SINGLELINE;
end;

procedure ColorizeBitmap(ABitmap: TBitmap; Color : TColor);
var
  i,j,l,Col: integer;
  R,G,B : Integer;
begin
  Color := ColorToRGB(Color);
  for i := 0 to ABitmap.Height - 1 do
    for j := 0 to ABitmap.Width - 1 do
      begin
        Col := ColorToRGB(ABitmap.Canvas.Pixels[j, i]);

        l := Round(((Col shr 16) + ((Col shr 8) and $00FF) + (Col and $0000FF)) div 3);

        R := GetRValue(Color) + l;
        G := GetGValue(Color) + l;
        B := GetBValue(Color) + l;
        if R > 255 then R := 255;
        if G > 255 then G := 255;
        if B > 255 then B := 255;
        ABitmap.Canvas.Pixels[j, i] := RGB(R, G, B);
      end;
end;

procedure AntialiasBitmap(ABitmap1, ABitmap2: TBitmap; Sampling: integer);
var
  yDest, xDest, i, j, R, G, B : integer;
  rowDest, rowSrc : PRGBArray;
begin
  for yDest:= 0 to ABitmap2.Height -1 do
  begin
    rowDest := ABitmap2.ScanLine[yDest];
    rowSrc := ABitmap1.ScanLine[yDest*Sampling];
    for xDest :=0 to ABitmap2.Width-1 do
    begin
      R:=0; G:=0; B:=0;
      // Make the sampling on Src image
      for i:=0 to Sampling-1 do
      begin
        for j:=0 to Sampling-1 do
        begin
          R := R + rowSrc[(xDest*Sampling)+j].rgbtRed;
          G := G + rowSrc[(xDest*Sampling)+j].rgbtGreen;
          B := B + rowSrc[(xDest*Sampling)+j].rgbtBlue;
        end;
      rowSrc := ABitmap1.ScanLine[(yDest*Sampling)+i];
      end;
      // Assign to the result image
      rowDest[xDest].rgbtRed := R div Sqr(Sampling);
      rowDest[xDest].rgbtGreen := G div Sqr(Sampling);
      rowDest[xDest].rgbtBlue := B div Sqr(Sampling);
    end;
  end;
end;

procedure AntialiasShapeBitmap(ABitmap1, ABitmap2: TBitmap; Sampling: integer; CornerWidth: integer; InnerColor: TColor);
var
  yDest, xDest, i, j, R, G, B : integer;
  rowDest, rowSrc : PRGBArray;
begin
  InnerColor := ColorToRGB(InnerColor);

  if CornerWidth < 1 then CornerWidth := Sampling;
  CornerWidth := CornerWidth div 2;

  try
    for yDest:= 0 to ABitmap2.Height - 1 do
    begin
      rowDest := ABitmap2.ScanLine[yDest];
      rowSrc := ABitmap1.ScanLine[yDest*Sampling];
      for xDest :=0 to ABitmap2.Width - 1 do
      begin
        R:=0; G:=0; B:=0;
        // Make the sampling on Src image
        if (xDest <= CornerWidth) or (yDest <= CornerWidth) or
           (xDest >= ABitmap2.Width - CornerWidth) or (yDest >= ABitmap2.Height - CornerWidth) then begin
          for i:=0 to Sampling-1 do
          begin
            for j:=0 to Sampling-1 do
            begin
              R := R + rowSrc[(xDest*Sampling)+j].rgbtRed;
              G := G + rowSrc[(xDest*Sampling)+j].rgbtGreen;
              B := B + rowSrc[(xDest*Sampling)+j].rgbtBlue;
            end;
          rowSrc := ABitmap1.ScanLine[(yDest*Sampling)+i];

          end;
          // Assign to the result image
          rowDest[xDest].rgbtRed := R div Sqr(Sampling);
          rowDest[xDest].rgbtGreen := G div Sqr(Sampling);
          rowDest[xDest].rgbtBlue := B div Sqr(Sampling);
        end;
      end;
    end;
  except
  end;

  ABitmap2.Canvas.Brush.Color := InnerColor;
  ABitmap2.Canvas.FillRect(Rect(CornerWidth, CornerWidth, ABitmap2.Width - CornerWidth, ABitmap2.Height - CornerWidth));
end;

procedure RbAlphaBlend(ABitmap1, ABitmap2: TBitmap; ARect: TRect; percent : integer);
var
  yDest, xDest: integer;
  rowDest, rowSrc : PRGBArray;
  WeightA, WeightB: Double;
begin
  WeightA := (100 - percent) / 100;
  WeightB := 1- WeightA;
  for yDest := ARect.Top to ARect.Bottom - 1 do
  begin
    rowDest := ABitmap1.ScanLine[yDest];
    rowSrc := ABitmap2.ScanLine[yDest - ARect.Top];
    for xDest := ARect.Left to ARect.Right - 1 do
    begin
      // Assign to the result image
      rowDest[xDest].rgbtRed := Round((rowDest[xDest].rgbtRed * WeightA + rowSrc[xDest - ARect.Left].rgbtRed * WeightB));
      rowDest[xDest].rgbtGreen := Round((rowDest[xDest].rgbtGreen * WeightA + rowSrc[xDest - ARect.Left].rgbtGreen * WeightB));
      rowDest[xDest].rgbtBlue := Round((rowDest[xDest].rgbtBlue * WeightA + rowSrc[xDest - ARect.Left].rgbtBlue * WeightB));
    end;
  end;
end;

{---------------------------------------------
---------------------------------------------}

procedure TShCustomControl.CMEnabledChanged(var Message: TMessage);
begin
  EventEnabledChanged;
  inherited;
end;

procedure TShCustomControl.CMFocusChanged(var Message: TMessage);
begin
  EventFocusChanged;
  inherited;
end;

procedure TShCustomControl.CMMouseEnter(var Message: TMessage);
begin
  EventMouseEnter;
  inherited;
end;

procedure TShCustomControl.CMMouseLeave(var Message: TMessage);
begin
  EventMouseLeave;
  inherited;
end;

procedure TShCustomControl.CMParentColorChanged(var Message: TMessage);
begin
  EventParentColorChanged;
  inherited;
end;

procedure TShCustomControl.CMTextChanged(var Message: TMessage);
begin
  EventTextChanged;
  inherited;
end;

constructor TShCustomControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque, csReplicatable] - [csDoubleClicks];
  DoubleBuffered := True;
  TabStop := True;
  FTextShadow := true;
  FShowCaption := true;
end;

destructor TShCustomControl.Destroy;
begin
  inherited;
end;

procedure TShCustomControl.DoBlendMore(Sender: TObject);
begin
  if (dsClicking in FDrawStates) and (dsMouseOver in FDrawStates) then
  begin
    if FClickedBlendPercent >= 100 then
    begin
      FClickedBlendPercent := 100;
      Invalidate;
      KillTimer(Handle, 1);
      Exit;
    end else
    begin
      Inc(FClickedBlendPercent, ClickBlendInterval);
    end;
  end else
  begin
    if FClickedBlendPercent > 0 then Dec(FClickedBlendPercent, ClickBlendInterval);

    if FOverBlendPercent >= 100 then begin
      FOverBlendPercent := 100;
    end else
     Inc(FOverBlendPercent, OverBlendInterval);
    if (FClickedBlendPercent <= 0) and (FOverBlendPercent >= 100) then begin
       Invalidate;
       KillTimer(Handle, 1);
       Exit;
    end;
  end;
  Invalidate;
end;

procedure TShCustomControl.DoBlendLess(Sender: TObject);
begin
  if FClickedBlendPercent > 0 then Dec(FClickedBlendPercent, ClickBlendInterval);
  if (FOverBlendPercent <= 0) and (FClickedBlendPercent <= 0) then
  begin
    FOverBlendPercent := 0;
    FClickedBlendPercent := 0;
    KillTimer(Handle, 1);
  end else
    Dec(FOverBlendPercent, OverBlendInterval);
  Invalidate;
end;

procedure TShCustomControl.EventEnabledChanged;
begin
  Invalidate;
end;

procedure TShCustomControl.EventFocusChanged;
begin
  ;
end;

procedure TShCustomControl.EventMouseDown;
begin
  if not Focused and CanFocus then
    SetFocus;
end;

procedure TShCustomControl.EventMouseEnter;
begin
  if Assigned(FOnMouseEnter) then
    FOnMouseEnter(Self);
end;

procedure TShCustomControl.EventMouseLeave;
begin
  if Assigned(FOnMouseLeave) then
    FOnMouseLeave(Self);
end;

procedure TShCustomControl.EventMouseMove;
begin
  ;
end;

procedure TShCustomControl.EventMouseUp;
begin
  ;
end;

procedure TShCustomControl.EventParentColorChanged;
begin
  ;
end;

procedure TShCustomControl.EventResized;
begin
  ;
end;

procedure TShCustomControl.EventTextChanged;
begin
  Invalidate;
end;

procedure TShCustomControl.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if Button = mbLeft then EventMouseDown;
end;

procedure TShCustomControl.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if Button = mbLeft then EventMouseUp;
end;

procedure TShCustomControl.SetTextShadow(const Value: boolean);
begin
  if FTextShadow <> Value then begin
    FTextShadow := Value;
    Invalidate;
  end;
end;

procedure TShCustomControl.WMMouseMove(var Message: TWMMouse);
begin
  EventMouseMove;
  inherited;
end;

procedure TShCustomControl.WMSize(var Message: TWMSize);
begin
  EventResized;
  inherited;
end;

procedure TShCustomControl.UpdateTimer;
begin
  KillTimer(Handle, 1);
  SetTimer(Handle, 1, AFadeSpeed[FFadeSpeed], nil);
end;

procedure TShCustomControl.WndProc(var Msg: TMessage);
begin
  with Msg do
    if Msg = WM_TIMER then
      try
        FOnFade(Self);
      except
      end;
  inherited;
end;

procedure TShCustomControl.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
end;

procedure TShCustomControl.SetShowCaption(const Value: boolean);
begin
  if FShowCaption <> Value then begin
    FShowCaption := Value;
    Invalidate;
  end;
end;

{---------------------------------------------
              TShControlColors
---------------------------------------------}

procedure TShControlColors.AssignTo(Dest: TPersistent);
begin
  if (Dest is TShControlColors) or Dest.InheritsFrom(TShControlColors) then
    with TShControlColors(Dest) do begin
      FDefaultFrom := Self.FDefaultFrom;
      FDefaultTo := Self.FDefaultTo;
      FOverFrom := Self.FOverFrom;
      FOverTo := Self.FOverTo;
      FClickedFrom := Self.FClickedFrom;
      FClickedTo := Self.FClickedTo;
      FBorderColor := Self.FBorderColor;
      FHotTrack := Self.FHotTrack;
      FTextShadow := Self.FTextShadow;
    end
  else
    inherited AssignTo(Dest);
end;

constructor TShControlColors.Create(AOwner: TComponent);
begin
  inherited;
  FOwner := AOwner;

  FClickedFrom := $00C6CFD6;
  FClickedTo   := $00EBF3F7;
  FHotTrack    := clBlue;
  FBorderColor := clGray;
  FTextShadow  := clWhite;
end;

procedure TShControlColors.DoUpdateGradients;
begin
  if FOwner <> nil then
    TShCustomControl(FOwner).UpdateGradients;
end;

procedure TShControlColors.SetBorderColor(const Value: TColor);
begin
  if FBorderColor <> Value then begin
    FBorderColor := Value;
    DoUpdateGradients;
  end;
end;

procedure TShControlColors.SetClickedFrom(const Value: TColor);
begin
  if Value <> FClickedFrom then begin
    FClickedFrom := Value;
    DoUpdateGradients;
  end;
end;

procedure TShControlColors.SetClickedTo(const Value: TColor);
begin
 if FClickedTo <> Value then begin
   FClickedTo := Value;
   DoUpdateGradients;
 end;
end;

procedure TShControlColors.SetHotTrack(const Value: TColor);
begin
  if FHotTrack <> Value then begin
    FHotTrack := Value;
    if FOwner <> nil then
      TCustomControl(FOwner).Invalidate;
  end;
end;

procedure TShControlColors.SetTextShadow(const Value: TColor);
begin
  if FTextShadow <> Value then begin
    FTextShadow := Value;
    if FOwner <> nil then
      TCustomControl(FOwner).Invalidate;
  end;
end;

{---------------------------------------------
             TShCustomColors
---------------------------------------------}

procedure TShCustomColors.AssignTo(Dest: TPersistent);
begin
  if Dest.InheritsFrom(TShCustomColors) or (Dest is TShCustomColors) then
    with TShCustomColors(Dest) do begin
      FDefaultFrom := Self.FDefaultFrom;
      FDefaultTo := Self.FDefaultTo;
      FOverFrom := Self.FOverFrom;
      FOverTo := Self.FOverTo;
    end
  else
    inherited AssignTo(Dest);
end;

constructor TShCustomColors.Create(AOwner: TComponent);
begin
  FOwner := AOwner;

  FDefaultFrom := clWhite;
  FDefaultTo   := $00D1BEAF;
  FOverFrom    := clWhite;
  FOverTo      := $00BE9476;
end;

procedure TShCustomColors.SetDefaultFrom(const Value: TColor);
begin
  if FDefaultFrom <> Value then begin
    FDefaultFrom := Value;
    DoUpdateGradients;
 end;
end;

procedure TShCustomColors.SetDefaultTo(const Value: TColor);
begin
  if FDefaultTo <> Value then begin
    FDefaultTo := Value;
    DoUpdateGradients;
 end;
end;

procedure TShCustomColors.SetOverFrom(const Value: TColor);
begin
  if FOverFrom <> Value then begin
    FOverFrom := Value;
    DoUpdateGradients;
 end;
end;

procedure TShCustomColors.SetOverTo(const Value: TColor);
begin
 if FOverTo <> Value then begin
    FOverTo := Value;
    DoUpdateGradients;
 end;
end;

{---------------------------------------------
              TShButton
---------------------------------------------}

constructor TShButton.Create(AOwner: TComponent);
begin
  inherited;
  Height := 25;
  Width := 75;
  BtnStyle:=bsNormal;

  FGrBitmap := TBitmap.Create;
  FGrBitmap.PixelFormat := pf24bit;
  FGrBitmap.Width := Width;
  FGrBitmap.Height := Height;

  FGrMouseOver := TBitmap.Create;
  FGrMouseOver.PixelFormat := pf24bit;
  FGrMouseOver.Width := Width;
  FGrMouseOver.Height := Height;

  FGrClicked := TBitmap.Create;
  FGrClicked.PixelFormat := pf24bit;
  FGrClicked.Width := Width;
  FGrClicked.Height := Height;

  FGradientType := gtVertical;

  FGlyph := TBitmap.Create;
  FGlyph.Transparent := true;

  FOverBlendPercent := 0;
  FClickedBlendPercent := 0;
  FDrawStates := [];
  FSpacing := 2;
  FLayout := blGlyphLeft;
  FadeSpeed := fsMedium;
  FShowFocusRect := true;
  FHotTrack := false;
  FGradientBorder := true;

  FDown := False;
  FAllowAllUp := false;
  FGroupIndex := 0;

  FColors := TShControlColors.Create(Self);
end;

destructor TShButton.Destroy;
begin
  FGrBitmap.Free;
  FGrMouseOver.Free;
  FGrClicked.free;
  FGlyph.Free;
  FColors.Free;
  inherited;
end;

procedure TShButton.EventResized;
begin
  inherited;
  UpdateGradients;
end;

procedure TShButton.EventMouseEnter;
begin
  inherited;
  Include(FDrawStates, dsMouseOver);
  FOnFade := DoBlendMore;
  UpdateTimer;
end;

procedure TShButton.EventMouseLeave;
begin
  inherited;
  if not FDown then begin
    Exclude(FDrawStates, dsMouseOver);
    FOnFade := DoBlendLess;
    UpdateTimer;
  end;
end;

procedure TShButton.EventMouseDown;
begin
  inherited;
  Include(FDrawStates, dsClicking);
  Include(FDrawStates, dsMouseOver);
  FOnFade := DoBlendMore;
  UpdateTimer;
end;

procedure TShButton.EventMouseUp;
begin
  inherited;
  if (dsClicking in FDrawStates) and not FDown then
    Exclude(FDrawStates, dsClicking);
  if dsMouseOver in FDrawStates then
    FOnFade := DoBlendMore;
  UpdateTimer;
end;

procedure TShButton.SetGlyph(const Value: TBitmap);
begin
  FGlyph.Assign(Value);
  FGlyph.Transparent := true;
  Invalidate;
end;

procedure TShButton.SetSpacing(const Value: integer);
begin
  if Value < 0 then
    FSpacing := 0
  else
    FSpacing := Value;
end;

procedure TShButton.SetLayout(const Value: TButtonLayout);
begin
  if Value = FLayout then Exit;
  FLayout := Value;
  Invalidate;
end;

procedure TShButton.SetFadeSpeed(const Value: TFadeSpeed);
begin
  FFadeSpeed := Value;
end;

procedure TShButton.SetHotTrack(const Value: boolean);
begin
  if FHotTrack <> Value then begin
    FHotTrack := Value;
    Invalidate;
  end;
end;


procedure TShButton.SetFocusRect(const Value: boolean);
begin
  if FShowFocusRect <> Value then begin
    FShowFocusRect := Value;
    Invalidate;
  end;
end;

procedure TShButton.UpdateGradients;
var
  R : TRect;
begin
  FGrBitmap.Width := Width;
  FGrBitmap.Height := Height;
  R := FGrBitmap.Canvas.ClipRect;
  DoBitmapGradient(FGrBitmap, R, Colors.DefaultFrom, Colors.DefaultTo, GradientType, 0);

  FGrMouseOver.Width := Width;
  FGrMouseOver.Height := Height;
  DoBitmapGradient(FGrMouseOver, R, Colors.OverFrom, Colors.OverTo, GradientType, 0);

  FGrClicked.Width := Width;
  FGrClicked.Height := Height;
  DoBitmapGradient(FGrClicked, R, Colors.ClickedFrom, Colors.ClickedTo, GradientType, 0);
  Invalidate;
end;

procedure TShButton.SetDefault(const Value: boolean);
begin
  if FDefault <> Value then begin
    FDefault := Value;
      with GetParentForm(Self) do
        Perform(CM_FOCUSCHANGED, 0, LongInt(ActiveControl));
  end;
end;

procedure TShButton.CMDialogKey(var Message: TCMDialogKey);
begin
  inherited;
  with Message do
    if (((CharCode = VK_RETURN) and (Focused or FDefault))
      or ((CharCode = VK_ESCAPE) and FCancel) and (KeyDataToShiftState(KeyData) = []))
      and CanFocus then
    begin
      Click;
      Result := 1;
    end
    else
      inherited;
end;

procedure TShButton.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if (Shift = []) and (Key = VK_SPACE) then Click;
  inherited;
end;

procedure TShButton.Click;
var
  Form: TCustomForm;
begin
  Form := GetParentForm(Self);
  if Form <> nil then
    Form.ModalResult := ModalResult;
  Down := (GroupIndex > 0) and not Down;
  inherited Click;
end;

procedure TShButton.EventFocusChanged;
begin
  inherited;
  Invalidate;
end;

procedure TShButton.SetAllowAllUp(const Value: boolean);
begin
  FAllowAllUp := Value;
end;

procedure TShButton.SetDown(const Value: boolean);
begin
  if (GroupIndex > 0) then begin
    if not Value and not CanbeUp then Exit;
    FDown := Value;
    if FDown then begin
      DeactivateSiblings;
      Include(FDrawStates, dsMouseOver);
      Include(FDrawStates, dsClicking);
      FOnFade := DoBlendMore;
      UpdateTimer;
    end else begin
      Exclude(FDrawStates, dsClicking);
      FOnFade := DoBlendLess;
      UpdateTimer;
    end;
  end;
end;

procedure TShButton.SetGroupIndex(const Value: integer);
begin
  FGroupIndex := Value;
end;

procedure TShButton.DeActivateSiblings;
var
  i : integer;
begin
  if (GroupIndex > 0) and (Parent <> nil) then
    with Parent do begin
      for i := 0 to ControlCount - 1 do
        if (Controls[i] <> Self) and (Controls[i] is TShButton) and (TShButton(Controls[i]).GroupIndex = GroupIndex) then
          TShButton(Controls[i]).SetDown(false);
    end;
end;

function TShButton.CanBeUp: boolean;
var
  i : integer;
begin
  Result := false;
  if AllowAllUp then begin
    Result := true;
    Exit;
  end;
  if (GroupIndex > 0) and (Parent <> nil) then
    with Parent do begin
      for i := 0 to ControlCount - 1 do
        if (Controls[i] <> Self) and (Controls[i] is TShButton) and (TShButton(Controls[i]).GroupIndex = GroupIndex) then
          if TShButton(Controls[i]).Down then begin
            Result := true;
            Exit;
          end;
    end;
end;

procedure TShButton.SetGradientType(const Value: TGradientType);
begin
  if FGradientType <> Value then begin
    FGradientType := Value;
    UpdateGradients;
  end;
end;

procedure TShButton.SetBtnStyle(const Value: TBtnStyle);
begin
  if FBtnStyle <> Value then FBtnStyle := Value;
end;

procedure TShButton.Paint;
var
  R : TRect;
  Flags: integer;
  FinalBitmap: TBitmap;
  Gl : TBitmap;
begin
  inherited;
  FinalBitmap := TBitmap.Create;
  try
    FinalBitmap.PixelFormat := pf24bit;
    FinalBitmap.Width := Width;
    FinalBitmap.Height := Height;

    R := GetClientRect;

    //Blend the base gradient and the MouseOver Gradient
    BlendBitmaps(FinalBitmap, FGrBitmap, FGrMouseOver, FOverBlendPercent, R, 1);

    //Blend the previous gradient and the Selection Gradient
    if FClickedBlendPercent > 0 then begin
      if GradientBorder then
        InflateRect(R, -3, -3);

      BlendBitmaps(FinalBitmap, FinalBitmap, FGrClicked, FClickedBlendPercent, R, 1);

      if GradientBorder then
        InflateRect(R, 3, 3);
    end;

    FinalBitmap.Canvas.Brush.Style := bsClear;
    FinalBitmap.Canvas.Pen.Color := Color;
    FinalBitmap.Canvas.Rectangle(R.Left, R.Top, R.Right, R.Bottom);

    FinalBitmap.Canvas.Pen.Color := FColors.BorderColor;

    with FinalBitmap.Canvas do
      //bsRect,bsRT,bsRTRB,bsRB,bsLT,bsLTLB,bsLB
      case FBtnStyle of
        bsNormal : RoundRect(0, 0, Width, Height, 5, 5);
        bsRect:
              begin
                MoveTo(0, 0);
                LineTo(Width-1, 0);
                LineTo(Width-1, Height-1);
                LineTo(0, Height-1);
                LineTo(0, 0);
              end;
        bsRT:
              begin
                MoveTo(0, 0);
                LineTo(Width-3, 0);
                LineTo(Width-1, 2);
                LineTo(Width-1, Height-1);
                LineTo(0, Height-1);
                LineTo(0, 0);
              end;
        bsRTRB:
              begin
                MoveTo(0, 0);
                LineTo(Width-3, 0);
                LineTo(Width-1, 2);
                LineTo(Width-1, Height-3);
                LineTo(Width-3, Height-1);
                LineTo(0, Height-1);
                LineTo(0, 0);
              end;
        bsRB:
              begin
                MoveTo(0, 0);
                LineTo(Width-1, 0);
                LineTo(Width-1, Height-3);
                LineTo(Width-3, Height-1);
                LineTo(0, Height-1);
                LineTo(0, 0);
              end;
        bsLT:
              begin
                MoveTo(2, 0);
                LineTo(Width-1, 0);
                LineTo(Width-1, Height-1);
                LineTo(0, Height-1);
                LineTo(0, 2);
                LineTo(2, 0);
              end;
        bsLTLB:
              begin
                MoveTo(2, 0);
                LineTo(Width-1, 0);
                LineTo(Width-1, Height-1);
                LineTo(2, Height-1);
                LineTo(0, Height-3);
                LineTo(0, 2);
                LineTo(2, 0);
              end;
        bsLB:
              begin
                MoveTo(0, 0);
                LineTo(Width-1, 0);
                LineTo(Width-1, Height-1);
                LineTo(2, Height-1);
                LineTo(0, Height-3);
                LineTo(0, 0);
              end;
      end;

    BitBlt(Canvas.Handle, 0, 0, Width, Height, FinalBitmap.Canvas.Handle, 0, 0, SRCCOPY);
  finally
    FinalBitmap.Free;
  end;

  //Calc text rect
  if not Glyph.Empty then begin
    case FLayout of
      blGlyphLeft :
        Inc(R.Left, FGlyph.Width + FSpacing);
      blGlyphRight :
        Dec(R.Left, FGlyph.Width + FSpacing);
      blGlyphTop :
        Inc(R.Top, FGlyph.Width + FSpacing);
      blGlyphBottom :
        Dec(R.Top, FGlyph.Width + FSpacing);
    end;
  end;

  SetBkMode(Canvas.Handle, Windows.TRANSPARENT);
  SetTextFlags(taCenter, false, Flags);

  //Draw text
  Canvas.Font.Assign(Font);
  if FHotTrack and (dsMouseOver in FDrawStates) then Canvas.Font.Color := FColors.HotTrack;

  if ShowCaption then
    DoDrawText(Canvas, Caption, Canvas.Font, Enabled, false, R, Flags, TextShadow, Colors.TextShadow);

  //Draw the glyph
  if not Glyph.Empty then begin
    gl := TBitmap.Create;
    try
      gl.Assign(FGlyph);
      gl.PixelFormat := pf24bit;
      gl.Transparent := true;
      if not Enabled then
        ColorizeBitmap(gl, $00404040);

      case FLayout of
        blGlyphLeft :
          if (Caption <> '') and ShowCaption then
            Canvas.Draw((R.Left + R.Right - canvas.TextWidth(Caption)) div 2 - FSpacing - FGlyph.Width, (Height - FGlyph.Height) div 2, gl)
          else
            Canvas.Draw((R.Left + R.Right - FSpacing) div 2 - FGlyph.Width, (Height - FGlyph.Height) div 2, gl);

        blGlyphRight :
          if (Caption <> '') and ShowCaption then
            Canvas.Draw((R.Left + R.Right + canvas.TextWidth(Caption)) div 2 + FSpacing, (Height - FGlyph.Height) div 2, gl)
          else
            Canvas.Draw((R.Left + R.Right) div 2 + FSpacing, (Height - FGlyph.Height) div 2, gl);
        blGlyphTop :
          if (Caption <> '') and ShowCaption then
            Canvas.Draw((Width - FGlyph.Width) div 2, (R.Top + R.Bottom - canvas.TextHeight(Caption)) div 2 - FSpacing - FGlyph.Height,  gl)
          else
            Canvas.Draw((Width - FGlyph.Width) div 2, (R.Top + R.Bottom) div 2 - FSpacing - FGlyph.Height,  gl);
        blGlyphBottom :
          if (Caption <> '') and ShowCaption then
            Canvas.Draw((Width - FGlyph.Width) div 2, (R.Top + R.Bottom + canvas.TextHeight(Caption)) div 2 + FSpacing,  gl)
          else
            Canvas.Draw((Width - FGlyph.Width) div 2, (R.Top + R.Bottom) div 2 + FSpacing,  gl);
      end;
    finally
      gl.Free;
    end;
  end;

  R := GetClientRect;

  if Focused and FShowFocusRect then begin
    InflateRect(R, -2, -2);
    DrawFocusRect(Canvas.Handle, R);
  end;
end;

{---------------------------------------------
              TShPanel
---------------------------------------------}

constructor TShCustomPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  DoubleBuffered := True;

  Width           := 100;
  Height          := 100;
  fStep           := 30;
  fSlideMode      := smNone;
  fSlideStyle     := ssNormal;
  fMin            := 20;
  fWidthMax       := Width;
  fHeightMax      := Height;
  fKoeff          := round(fHeightMax*fStep/fWidthMax);

  fTimer          := TTimer.Create(Self);
  fTimer.Enabled  := false;
  fTimer.Interval := 2;
  fTimer.OnTimer  := TimerTimer;

  fGlyph := TBitmap.Create;
  fGlyph.Transparent := true;
  fGlyph.LoadFromResourceName(HInstance,'GLYPTH');

  fCollapse       := false;
  Color           := $e5dfdf; //f2ece9;// e9e0d9;
  fGradient       := gsNone;
end;

destructor TShCustomPanel.Destroy;
begin
  fTimer.Free;
  fGlyph.Free;
  inherited;
end;

procedure TShCustomPanel.TimerTimer(Sender: TObject);
var
  i: integer;
begin
  if (fSlideMode=smNone) or not Enabled then begin fTimer.Enabled:=false; Exit; end;
  if fSlideMode in [smTop,smBottom] then fKoeff:=fStep;
  if (fSlideStyle in [ssCombine]) or (fSlideMode in [smTop,smBottom]) then begin

    Height:=Height-fKoeff;
    if fSlideMode in [smBottom] then Top:=Top+fKoeff;

    if Height-fKoeff>=fHeightMax then begin
      if fSlideMode in [smBottom] then
        Top:=Top-(fHeightMax-Height);
      Height:=fHeightMax;
      if fSlideMode in [smTop,smBottom] then begin
        fTimer.Enabled:=false;
        fStep:=-fStep;
        fKoeff:=-fKoeff;
        fClosed:=false;
        Exit;
      end;
    end;
    if Height-fKoeff<=fMin then begin
      if (fSlideMode in [smBottom]) then
        Top:=Top+(Height-fMin);
      Height:=fMin;
      if fSlideMode in [smTop,smBottom] then begin
        fTimer.Enabled:=false;
        fStep:=-fStep;
        fKoeff:=-fKoeff;
        fClosed:=true;
        Exit;
      end;
    end;

  end;

  for i:=0 to ControlCount-1 do begin
    if fSlideMode in [smFull,smLeft] then
      Controls[i].Left:=Controls[i].Left-fStep;
    if fSlideMode in [smTop] then
      Controls[i].Top:=Controls[i].Top-fStep;
  end;

  if (Width-fStep>=fWidthMax) and (fStep<0) and not (fSlideMode in [smTop,smBottom]) then begin

    if Width-fStep>fWidthMax then
      case fSlideMode of
        smFull,
        smLeft:
                begin
                  Width:=fWidthMax;
                end;
        smRight:
                begin
                  Left:=Left-(fWidthMax-Width);
                  Width:=fWidthMax;
                end;
      end;

    fTimer.Enabled:=false;
    fStep:=-fStep;
    fKoeff:=-fKoeff;
    fClosed:=false;
    Exit;
  end;

  if (Width-fStep<=fMin) and (fStep>0) then begin

    if Width-fStep<fMin then
      case fSlideMode of
          smFull,
          smLeft:
                  begin
                    Width:=fMin;
                  end;
          smRight:
                  begin
                    Left:=Left+(Width-fMin);
                    Width:=fMin;
                  end;
      end;

    fTimer.Enabled:=false;
    fStep:=-fStep;
    fKoeff:=-fKoeff;
    fClosed:=true;
    Exit;
  end;

  case fSlideMode of
    smFull,
    smLeft: Width:=Width-fStep;
    smRight:
            begin
              Width:=Width-fStep;
              Left:=Left+fStep;
            end;
  end;

end;

procedure TShCustomPanel.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (fSlideMode in [smFull,smLeft,smRight,smTop,smBottom]) and fGlyphOn then
    fTimer.Enabled:=true;
  inherited MouseUp(Button, Shift, X, Y);
end;

procedure TShCustomPanel.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  R: TRect;
begin
  case fSlideMode of
    smLeft,
    smFull,
    smBottom: R:=Rect(Width-fGlyph.Width-4,2,Width-4,fGlyph.Height+2);
    smRight : R:=Rect(4,2,fGlyph.Width+4,fGlyph.Height+2);
    smTop   : R:=Rect(Width-fGlyph.Width-4,Height-fGlyph.Height-2,Width-4,Height-2);
  end;
  fGlyphOn:=false;
  if  (X>R.Left) and (X<R.Right)
  and (Y>R.Top)  and (Y<R.Bottom) then begin
    Cursor:=crHandPoint;
    fGlyphOn:=true;
  end else
    Cursor:=crDefault;
  inherited MouseMove(Shift, X, Y);
end;

procedure TShCustomPanel.WMSize(var Message: TWMSize);
var
  Rgn1,Rgn2 : HRGN;
begin
  if HandleAllocated then begin
    Rgn1 := CreateRoundRectRgn(0,0,Width,Height,9,9);
    Rgn2:=0;
    case fSlideMode of
      smLeft   : Rgn2 := CreateRectRgn(0,0,9,Height-1);
      smRight  : Rgn2 := CreateRectRgn(Width-9,0,Width-1,Height-1);
      smTop    : Rgn2 := CreateRectRgn(0,0,Width-1,9);
      smBottom : Rgn2 := CreateRectRgn(0,Height-9,Width-1,Height-1);
    end;
    try
      if fSlideMode<>smNone then CombineRgn(Rgn1, Rgn1, Rgn2, RGN_OR);
      SetWindowRgn (Handle, Rgn1, TRUE);
    finally
      DeleteObject (Rgn1);
    end;
  end;
  inherited;
end;

procedure TShCustomPanel.SetGlyph(const Value: TBitmap);
begin
  fGlyph.Assign(Value);
  fGlyph.Transparent := true;
  Invalidate;
end;

procedure TShCustomPanel.SetMinMax(Min,WMax,HMax: integer);
begin
  fMin            := Min;
  fWidthMax       := WMax;
  fHeightMax      := HMax;
  fKoeff          := round(fHeightMax*fStep/fWidthMax);
end;

procedure TShCustomPanel.SetCollapse(const Value: boolean);
var
  i: integer;
begin
  if fCollapse<>Value then begin
    if Value then begin
      for i:=0 to ControlCount-1 do begin
        if fSlideMode in [smFull,smLeft] then
          Controls[i].Left:=Controls[i].Left-fWidthMax;
        if fSlideMode in [smTop] then
          Controls[i].Top:=Controls[i].Top-fHeightMax;
      end;
      case fSlideMode of
        smFull,
        smLeft,
        smRight  : Width:=fMin;
        smTop,
        smBottom : Height:=fMin;
      end;
      if (fSlideStyle=ssCombine) and (fSlideMode<>smNone) then Height:=fMin;
      fStep:=-fStep;
      fKoeff:=-fKoeff;
      fClosed:=true;
      Invalidate;
    end;
    fCollapse:=Value;
  end;
end;

procedure TShCustomPanel.Paint;
var
  fBuffer  : TBitmap;
  R        : TRect;
  ACanvas  : TCanvas;
  gl       : TBitmap;

  procedure ColorizeBitmap(ABitmap: TBitmap; Color : TColor);
  var
    i,j,l,Col: integer;
    R,G,B : Integer;
  begin
    Color := ColorToRGB(Color);
    for i := 0 to ABitmap.Height - 1 do
      for j := 0 to ABitmap.Width - 1 do
        begin
          Col := ColorToRGB(ABitmap.Canvas.Pixels[j, i]);
          l := Round(((Col shr 16) + ((Col shr 8) and $00FF) + (Col and $0000FF)) div 3);
          R := GetRValue(Color) + l;
          G := GetGValue(Color) + l;
          B := GetBValue(Color) + l;
          if R > 255 then R := 255;
          if G > 255 then G := 255;
          if B > 255 then B := 255;
          ABitmap.Canvas.Pixels[j, i] := RGB(R, G, B);
        end;
  end;

begin
  inherited;

  R               := Canvas.ClipRect;
  fBuffer         := TBitmap.Create;
  fBuffer.Width   := R.Right  - R.Left- 1;
  fBuffer.Height  := R.Bottom - R.Top - 1;
  ACanvas         := fBuffer.Canvas;
  ACanvas.Font    := Font;

  ACanvas.Brush.Style:=bsSolid;
  ACanvas.Brush.Color:=Color;
  PatBlt(ACanvas.Handle,0,0,Width,Height,PATCOPY);

  with ACanvas do begin

    case fGradient of
      gsLeftVertical     : DrawGradient(ACanvas,$ffffff,$e5dfdf,128,ACanvas.ClipRect,true);
      gsRightVertical    : DrawGradient(ACanvas,$e5dfdf,$ffffff,128,ACanvas.ClipRect,true);
      gsTopHorizontal    : DrawGradient(ACanvas,$88ccff,$eeffff,128,ACanvas.ClipRect,true);
      gsBottomHorizontal : DrawGradient(ACanvas,$eeffff,$88ccff,128,ACanvas.ClipRect,true);
    end;

    Brush.Style := bsClear;
    Pen.Style   := psSolid;
    Pen.Color   := clGray; 
    if fSlideMode in [smLeft,smRight,smTop,smBottom] then begin
      case fSlideMode of
        smLeft  : RoundRect(-10,0,Width-1,Height-1,9,9);
        smRight : RoundRect(0,0,Width+10,Height-1,9,9);
        smTop   : RoundRect(0,-10,Width-1,Height-1,9,9);
        smBottom: RoundRect(0,0,Width-1,Height+10,9,9);
      end;
      Rectangle(0,0,Width-1,Height-1);
    end else
      RoundRect(0,0,Width-1,Height-1,9,9);

    gl := TBitmap.Create;
    try
      gl.Assign(fGlyph);
      gl.PixelFormat := pf24bit;  
      gl.Transparent := true;
      if not Enabled then ColorizeBitmap(gl, $00404040);
      if fSlideMode in [smFull,smLeft,smRight,smTop,smBottom] then
        case fSlideMode of
          smLeft,
          smFull,
          smBottom: Draw(Width-gl.Width-4,2,gl);
          smRight : Draw(4,2,gl);
          smTop   : Draw(Width-gl.Width-4,Height-gl.Height-2,gl);
        end;
    finally
      gl.Free;
    end;

  end;

  Canvas.Draw(Canvas.ClipRect.Left, Canvas.ClipRect.Top, fBuffer);
  fBuffer.Free;

end;

end.
