unit ImageEditor;

interface

uses
  Windows, Messages, Classes, Graphics, Controls, ImgView, ImgDraw,
  GlowControls, SysUtils, Forms, SlideBar, ColorPicker, GR32;

const
  crSelCursor     = 20;
  crMoveCursor    = 21;

  crFillCursor    = 23;
  crBrushCursor   = 24;
  crPenCursor     = 25;
  crRectCursor    = 26;
  crFillRectCur   = 27;
  crERectCursor   = 28;
  crEFillRectCur  = 29;
  crElipCursor    = 30;

  crPipeCursor    = 32;
  crLineCursor    = 33;
  crPolyCursor    = 34;
  crSelInCursor   = 35;

type
  TCustomEditor    = class;

  TPropertyPanel    = class(TShPanel)
    private

    public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
  end;

  TColorPicker    = class(TShPanel)
    private
      fPicker       : TShColorPicker;
      fControl      : TCustomEditor;
      procedure     OnChange(Sender: TObject);
    public
      constructor   Create(AOwner: TComponent); override;
      destructor    Destroy; override;
  end;

  TSelectColor    = class(TShPanel)
    private
      fControl    : TCustomEditor;
      fForeColor  : TShPanel;
      fBackColor  : TShPanel;
      procedure   MouseClick(Sender: TObject);
      procedure   MouseDblClick(Sender: TObject);
    public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
  end;

  TToolPanel    = class(TShPanel)
    private
      fButtons     : TShButton;
      fControl     : TCustomEditor;
      fSelectColor : TSelectColor;
    public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
  end;

  TAnimatePanel    = class(TShPanel)
    private
      //fControl    : TCustomEditor;
    public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
  end;

  TImageView    = class(TShPanel)
    private
      fControl    : TCustomEditor;
      fViewControl: TCustomImageView;
    protected
      procedure   WMSize(var Message: TMessage); message WM_SIZE;
    public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
  end;

  TCustomEditor    = class(TCustomControl)
    private
      fToolPanel  : TToolPanel;
      fColorPicker: TColorPicker;
      fPropPanel  : TPropertyPanel;
      fImgView    : TImageView;
      fAnimate    : TAnimatePanel;
      fActivePanel: integer;
      fLoading    : boolean;
      procedure   OnPanelClick(Sender: TObject);
      procedure   ButtonClick(Sender: TObject);
    protected
      procedure   WMSize(var Message: TMessage); message WM_SIZE;
      procedure   CreateParams(var Params: TCreateParams); override;
    public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
  end;

  TImageEditor    = class(TCustomEditor)

  end;

implementation

{$R ImageEditor.RES}

//############################# TOOL ##################################

constructor TPropertyPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SlideMode:=smLeft;
  SlideStyle:=ssCombine;
  Width:=150;
  Height:=150;
  SetMinMax(22,Width,Height);
  Collapse:=true;
end;

destructor TPropertyPanel.Destroy;
begin
  inherited Destroy;
end;

//############################# COLORPIC ##################################

constructor TColorPicker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fControl :=TCustomEditor(AOwner);

  SlideMode:=smLeft;

  fPicker:=TShColorPicker.Create(Self);
  fPicker.Parent:=Self;
  fPicker.Width:=190;
  fPicker.Height:=160;
  fPicker.Top:=2;
  fPicker.Left:=5;
  fPicker.fColorPicker.OnChange:=OnChange;

  Width:=210;
  Height:=200;
  SetMinMax(22,Width,Height);
  Collapse:=true;
end;

destructor TColorPicker.Destroy;
begin
  fPicker.Free;
  inherited Destroy;
end;

procedure TColorPicker.OnChange(Sender: TObject);
begin
  case fControl.fActivePanel of
    0: fControl.fToolPanel.fSelectColor.fForeColor.Color:=fPicker.fColorPicker.SelectedColor;
    1: fControl.fToolPanel.fSelectColor.fBackColor.Color:=fPicker.fColorPicker.SelectedColor;
  end;
  fControl.fImgView.fViewControl.ActiveColor:=Color32(fControl.fToolPanel.fSelectColor.fForeColor.Color);
end;

//############################# COLOR ##################################

constructor TSelectColor.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fControl    := TCustomEditor(AOwner);
  Width:=50;
  Height:=50;
  Gradient:=gsLeftVertical;

  fBackColor  := TShPanel.Create(Self);
  fBackColor.Parent:=Self;
  fBackColor.Width:=27;
  fBackColor.Height:=27;
  fBackColor.Top:=20;
  fBackColor.Left:=20;
  fBackColor.Color:=clWhite;
  fBackColor.Tag:=1;
  fBackColor.OnClick:=MouseClick;
  fBackColor.OnDblClick:=MouseDblClick;

  fForeColor  := TShPanel.Create(Self);
  fForeColor.Parent:=Self;
  fForeColor.Width:=27;
  fForeColor.Height:=27;
  fForeColor.Top:=3;
  fForeColor.Left:=3;
  fForeColor.Color:=clBlack;
  fForeColor.Tag:=0;
  fForeColor.OnClick:=MouseClick;
  fForeColor.OnDblClick:=MouseDblClick;
end;

destructor TSelectColor.Destroy;
begin
  fForeColor.Free;
  fBackColor.Free;
  inherited Destroy;
end;

procedure TSelectColor.MouseClick(Sender: TObject);
begin
  fControl.fImgView.fViewControl.ActiveColor:=Color32(fForeColor.Color);
  if not fControl.fColorPicker.Close then begin
    fControl.fActivePanel:=TShPanel(Sender).Tag;
    case fControl.fActivePanel of
      0: fControl.fColorPicker.fPicker.SelectedColor:=fForeColor.Color;
      1: fControl.fColorPicker.fPicker.SelectedColor:=fBackColor.Color;
    end;
  end; 
end;

procedure TSelectColor.MouseDblClick(Sender: TObject);
begin
  if fControl.fColorPicker.Close then fControl.fColorPicker.fTimer.Enabled:=true;
  fControl.fActivePanel:=TShPanel(Sender).Tag;
  case fControl.fActivePanel of
    0: fControl.fColorPicker.fPicker.SelectedColor:=fForeColor.Color;
    1: fControl.fColorPicker.fPicker.SelectedColor:=fBackColor.Color;
  end;
end;

//############################# IMGVIEW ##################################

constructor TImageView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fControl    := TCustomEditor(AOwner);
  fViewControl:= TCustomImageView.Create(Self, Point(32,32));
  fViewControl.Parent:=Self;
  fViewControl.Left:=5;
  fViewControl.Top:=5;
  fViewControl.Color:=Color;
end;

destructor TImageView.Destroy;
begin
  fViewControl.Free;
  inherited Destroy;
end;

procedure TImageView.WMSIZE(var Message : TMessage);
begin
  inherited;
  fViewControl.Left:=7;
  fViewControl.Top:=7;
  fViewControl.Width:=Width-12;
  fViewControl.Height:=Height-12;
end;

//############################# ANIMATE ##################################

constructor TAnimatePanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SlideMode:=smBottom;
  SlideStyle:=ssNormal;
end;

destructor TAnimatePanel.Destroy;
begin
  inherited Destroy;
end;

//############################# TOOL ##################################

constructor TToolPanel.Create(AOwner: TComponent);
const
  Styles : array [Boolean] of TGradientType = (gtFromTopLeft,gtFromTopRight);
  BtnForm: array [Boolean] of TBtnStyle = (bsRTRB,bsLTLB);
var
  x,y,n : integer;
  Bmp   : TBitmap;
  Img   : TBitmap;
begin
  inherited Create(AOwner);
  fControl    := TCustomEditor(AOwner);
  Bmp:= TBitmap.Create;
  Bmp.LoadFromResourceName(HInstance,'TOOLBMP');
  Img:= TBitmap.Create;
  Img.Width :=24;
  Img.Height:=24;
  for y:=0 to 7 do begin
    for x:=0 to 1 do begin
      n:=(y+x)+y;
      fButtons:= TShButton.Create(self);
      fButtons.Width  := 25;
      fButtons.Height := 24;
      fButtons.Parent := Self;
      fButtons.Left   := x*24+5;
      fButtons.Top    := y*23+5;
      fButtons.Tag    := n;
      fButtons.GradientType   := Styles[x=0];
      fButtons.BtnStyle       := BtnForm[x=0];
      fButtons.ShowFocusRect  := false;
      fButtons.GroupIndex     := 1;
      fButtons.GradientBorder := false;
      fButtons.OnClick        := fControl.ButtonClick;
      BitBlt(Img.Canvas.Handle,0,0,24,24,Bmp.Canvas.Handle,n*25,0,SRCCOPY);
      fButtons.Glyph:=Img;
      Screen.Cursors[n+20] := LoadCursor(HInstance, PChar('CUR'+IntToStr(n)));
    end;
  end;
  Screen.Cursors[crSelInCursor] := LoadCursor(HInstance, 'CUR16');

  fSelectColor := TSelectColor.Create(AOwner);
  fSelectColor.Parent:=Self;
  fSelectColor.Left:=5;
  fSelectColor.Top:=fButtons.Top+fButtons.Height+5;

  Width:=60;
  Height:=fSelectColor.Top+fSelectColor.Height+5;

  Bmp.Free;
  Img.Free;

end;

destructor TToolPanel.Destroy;
begin
  fButtons.Free;
  fSelectColor.Free;
  inherited Destroy;
end;

//############################# EDITOR ##################################

constructor TCustomEditor.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  fToolPanel  := TToolPanel.Create(Self);
  fToolPanel.Parent:=Self;
  fToolPanel.Top:=5;
  fToolPanel.Left:=5;
  fToolPanel.Gradient:=gsRightVertical;

  fColorPicker:= TColorPicker.Create(Self);
  fColorPicker.Parent:=Self;
  fColorPicker.Left:=fToolPanel.Left+fToolPanel.Width-2;
  fColorPicker.Top:=fToolPanel.Height+fToolPanel.Top-fColorPicker.Height-5;
  fColorPicker.SendToBack;
  fColorPicker.OnClick:=OnPanelClick;

  fPropPanel      := TPropertyPanel.Create(Self);
  fPropPanel.Parent:=Self;
  fPropPanel.Left := fToolPanel.Left+fToolPanel.Width-2;
  fPropPanel.Top  := 28;
  fPropPanel.OnClick:=OnPanelClick;

  fImgView        := TImageView.Create(Self);
  fImgView.Parent := Self;
  fImgView.Left   := fToolPanel.Left+fToolPanel.Width+25;
  fImgView.Top    := 50;
  fImgView.Width  := 400;
  fImgView.Height := 400;
  fImgView.SendToBack;

  fAnimate        := TAnimatePanel.Create(Self);
  fAnimate.Parent := Self;
  fAnimate.Width  := 400;
  fAnimate.Height := 100;
  fAnimate.SetMinMax(22,fAnimate.Width,fAnimate.Height);
  fAnimate.Collapse:=true;
  fAnimate.Enabled:=false;
  fAnimate.Left   := 0;

  DoubleBuffered  := true;
  fLoading        := false;
  Color           := clWhite;
  fActivePanel    := 0;
end;

destructor TCustomEditor.Destroy;
begin
  fToolPanel.Free;
  fColorPicker.Free;
  fPropPanel.Free;
  fImgView.Free;
  fAnimate.Free;
  inherited Destroy;
end;

procedure TCustomEditor.OnPanelClick(Sender: TObject);
begin
  TShPanel(Sender).BringToFront;
  fToolPanel.BringToFront;
end;

procedure TCustomEditor.ButtonClick(Sender: TObject);
begin
  fImgView.fViewControl.AddLayer(32,32);
  fImgView.fViewControl.ToolMode:=TToolMode(TShButton(Sender).Tag);
  fImgView.fViewControl.ActiveCursor := TShButton(Sender).Tag+20;
  case fImgView.fViewControl.ToolMode of
    _move: fImgView.fViewControl.ImgLayer[fImgView.fViewControl.ActiveLayer].Lock:=true;
  else
    fImgView.fViewControl.ImgLayer[fImgView.fViewControl.ActiveLayer].Lock:=false;
  end;
  fImgView.fViewControl.ActiveColor:=Color32(fToolPanel.fSelectColor.fForeColor.Color);
end;

procedure TCustomEditor.WMSIZE(var Message : TMessage);
begin
  if fAnimate.Close then
    fAnimate.Top := Height-fAnimate.fMin+2
  else
    fAnimate.Top := Height-fAnimate.fHeightMax+2;
end;

procedure TCustomEditor.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
    WindowClass.style := WindowClass.style and not (CS_HREDRAW or CS_VREDRAW);
end;

end.
