unit ImgView;

interface

uses
  Windows, Messages, Classes, Controls, Graphics, ExtCtrls, Forms, ImgDraw,
  GR32_Image, GR32, GR32_Layers, SysUtils, Contnrs;

type
  TScrollBarInc     = 1..32767;

  TCustomImageView    = class;

  TCustomScrollBar  = class(TPersistent)
    private
      fControl    : TCustomImageView;
      fKind       : TScrollBarKind;
      fIncrement  : TScrollBarInc;
      fCalcRange  : Integer;
      fPosition   : integer;
      fRange      : integer;
      fSize       : integer;
      fTracking   : boolean;
      fVisible    : boolean;
      fEnabled    : boolean;
      constructor Create(AControl: TCustomImageView; AKind: TScrollBarKind);
      procedure   Update(ControlSB, AssumeSB: Boolean);
      procedure   SetPosition(Value: Integer);
      procedure   SetRange(Value: Integer);
      procedure   SetSize(Value: Integer);
      procedure   SetVisible(Value: Boolean);
      procedure   SetEnabled(Value: Boolean);
      function    ControlSize(ControlSB, AssumeSB: Boolean): Integer;
      function    NeedsScrollBarVisible: Boolean;
      procedure   ScrollMessage(var Msg: TWMScroll);
    published
      property    Position : Integer read fPosition write SetPosition default 0;
      property    Range    : Integer read fRange write SetRange default 0;
      property    Size     : Integer read fSize write SetSize default 0;
      property    Tracking : Boolean read fTracking write fTracking default False;
      property    Visible  : Boolean read fVisible write SetVisible default True;
      property    Enabled  : Boolean read fEnabled write SetEnabled default True;
      property    Increment: TScrollBarInc read fIncrement write fIncrement default 8;
  end;

  TWorkBitmapLayer = class(TBitmapLayer)
    private
      fControl    : TCustomImageView;
      fLeft       : integer;
      fTop        : integer;
      fWidth      : integer;
      fHeight     : integer;
      fLock       : boolean;
      procedure   SetLeft(Value: integer);
      procedure   SetTop(Value: integer);
    protected
      CorrX,CorrY : integer;
    public
      constructor Create(OWner: TCustomImageView; Size: TPoint; ALayerCollection: TLayerCollection); reintroduce;
      destructor  Destroy; override;
    published
      property    Lock : boolean read fLock write fLock;
      property    Left : integer read fLeft write SetLeft;
      property    Top  : integer read fTop  write SetTop;
  end;

  TCustomImageView    = class(TCustomImage32)
    private
      fHorzScrollBar: TCustomScrollBar;
      fVertScrollBar: TCustomScrollBar;
      fScale        : integer;
      fActiveLayer  : integer;
      fLayers       : TObjectList;
      fTmpLayer     : TWorkBitmapLayer;
      fActiveColor  : TColor32;
      fActiveCursor : TCursor;
      fDrawMode     : TToolMode;
      procedure   SetHorzScrollBar(Value: TCustomScrollBar);
      procedure   SetVertScrollBar(Value: TCustomScrollBar);
      procedure   SetActiveColor(Value: TColor32);
      procedure   SetActiveLayer(Value: integer);
      procedure   UpdateScrollBars;
      procedure   ScrollBy(DeltaX, DeltaY: Integer);
      function    GetLayer(const LayerIndex: Integer): TWorkBitmapLayer;
    protected
      fWidth      : integer;
      fHeight     : integer;
      fDirect     : integer;
      Ox,Oy       : integer;
      OffsetX     : integer;
      OffsetY     : integer;
      fDown       : boolean;
      Mx,My       : integer;
      SDx,SDy     : integer;
      EDx,EDy     : integer;
      NDx,NDy     : integer;
      procedure   WMSize      (var Message: TWMSize);       message WM_SIZE;
      procedure   WMHScroll   (var Message: TWMHScroll);    message WM_HSCROLL;
      procedure   WMVScroll   (var Message: TWMVScroll);    message WM_VSCROLL;
      procedure   WMMouseWheel(var Message: TMessage);      message WM_MOUSEWHEEL;
      procedure   WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
      procedure   WMPaint     (var Message: TMessage);      message WM_PAINT;

      procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure   MouseMove(Shift: TShiftState; X, Y: Integer); override;
      procedure   MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      function    MouseInRect(X,Y: integer): boolean;
      function    MouseInRange(X,Y: integer): boolean;

      procedure   PaintStage(Sender: TObject; Buffer: TBitmap32; StageNum: Cardinal);
    public
      constructor Create(AOwner: TComponent; Size: TPoint); reintroduce;
      destructor  Destroy; override;
      property    ImgLayer[const LayerIndex: Integer]: TWorkBitmapLayer read GetLayer;
      procedure   AddLayer(Width,Height: integer);
      procedure   SaveToFile(FName: string);
    published
      property    HorzScrollBar: TCustomScrollBar read fHorzScrollBar write SetHorzScrollBar;
      property    VertScrollBar: TCustomScrollBar read fVertScrollBar write SetVertScrollBar;
      property    ActiveColor: TColor32 read fActiveColor write SetActiveColor;
      property    ActiveLayer: integer read fActiveLayer write SetActiveLayer;
      property    ActiveCursor: TCursor read fActiveCursor write fActiveCursor;
      property    ToolMode: TToolMode read fDrawMode write fDrawMode;
      property    Color;
  end;

implementation

{$Region ' TCustomScrollBar '}
constructor TCustomScrollBar.Create(AControl: TCustomImageView; AKind: TScrollBarKind);
begin
  inherited Create;
  fControl    := AControl;
  fKind       := AKind;
  Tracking    := true;
  Visible     := true;
  Enabled     := true;
  fIncrement  := 8;
end;

function TCustomScrollBar.ControlSize(ControlSB, AssumeSB: Boolean): Integer;
var
  BorderAdjust: Integer;

  function ScrollBarVisible(Code: Word): Boolean;
  var
    Style: Longint;
  begin
    Style := WS_HSCROLL;
    if Code = SB_VERT then Style := WS_VSCROLL;
    Result := GetWindowLong(FControl.Handle, GWL_STYLE) and Style <> 0;
  end;

  function Adjustment(Code, Metric: Word): Integer;
  begin
    Result := 0;
    if not ControlSB then
      if AssumeSB and not ScrollBarVisible(Code) then
        Result := -(GetSystemMetrics(Metric) - BorderAdjust)
      else if not AssumeSB and ScrollBarVisible(Code) then
        Result := GetSystemMetrics(Metric) - BorderAdjust;
  end;

begin
  BorderAdjust := Integer(GetWindowLong(FControl.Handle, GWL_STYLE) and
    (WS_BORDER or WS_THICKFRAME) <> 0);
  if fKind = sbVertical then
    Result := FControl.ClientHeight + Adjustment(SB_HORZ, SM_CXHSCROLL) else
    Result := FControl.ClientWidth + Adjustment(SB_VERT, SM_CYVSCROLL);
end;

function TCustomScrollBar.NeedsScrollBarVisible: Boolean;
begin
  Result := FRange > ControlSize(False, False);
end;

procedure TCustomScrollBar.Update(ControlSB, AssumeSB: Boolean);
var
  ScrollInfo: TScrollInfo;
  Code      : Word;
begin
  fCalcRange            := 0;
  Code                  := SB_HORZ;
  if fKind = sbVertical then begin
    Code                := SB_VERT;
  end;
  if Visible then begin
    fCalcRange := Range - ControlSize(ControlSB, AssumeSB);
    if fCalcRange < 0 then fCalcRange := 0;
  end;
  ScrollInfo.cbSize     := SizeOf(ScrollInfo);
  ScrollInfo.fMask      := SIF_ALL;
  ScrollInfo.nMin       := 0;
  ScrollInfo.nMax       := Range;
  ScrollInfo.nPage      := (ControlSize(ControlSB, AssumeSB) + 1);
  ScrollInfo.nPos       := fPosition;
  ScrollInfo.nTrackPos  := fPosition;
  SetScrollInfo(fControl.Handle, Code, ScrollInfo, True);
  if not Visible then ShowScrollBar(fControl.Handle, Code, Visible);
  if Enabled then
    EnableScrollBar(fControl.Handle,Code,ESB_ENABLE_BOTH)
  else begin
    EnableScrollBar(fControl.Handle,Code,ESB_DISABLE_BOTH);
    Exit;
  end;
  SetPosition(fPosition);
end;

procedure TCustomScrollBar.ScrollMessage(var Msg: TWMScroll);

  function GetRealScrollPosition: Integer;
  var
    SI: TScrollInfo;
    Code: Integer;
  begin
    SI.cbSize := SizeOf(TScrollInfo);
    SI.fMask := SIF_TRACKPOS;
    Code := SB_HORZ;
    if fKind = sbVertical then Code := SB_VERT;
    Result := Msg.Pos;
    if GetScrollInfo(FControl.Handle, Code, SI) then
      Result := SI.nTrackPos;
  end;

begin
  with Msg do
    case ScrollCode of
      SB_LINEUP       : SetPosition(Position - fIncrement);
      SB_LINEDOWN     : SetPosition(Position + fIncrement);
      SB_PAGEUP       : SetPosition(Position - ControlSize(True, False));
      SB_PAGEDOWN     : SetPosition(Position + ControlSize(True, False));
      SB_TOP          : SetPosition(0);
      SB_BOTTOM       : SetPosition(fCalcRange);
      SB_THUMBPOSITION:
                        if fCalcRange > 32767 then
                          SetPosition(GetRealScrollPosition)
                        else
                          SetPosition((Pos div Increment)*Increment);
      SB_THUMBTRACK   :
                        if Tracking then
                          if fCalcRange > 32767 then
                            SetPosition(GetRealScrollPosition)
                          else
                            SetPosition((Pos div Increment)*Increment);
      SB_ENDSCROLL    : ;
    end;
end;

procedure TCustomScrollBar.SetPosition(Value: Integer);
var
  Code    : Word;
  OldPos  : Integer;
begin
  if csReading in fControl.ComponentState then begin
    fPosition := (Value div fIncrement)*fIncrement;
    Exit;
  end;
  
  if not fControl.HandleAllocated or (Value=fPosition) then Exit;

  if Value > fCalcRange then
    Value := fCalcRange
  else if Value < 0 then
    Value := 0;

  if fKind = sbHorizontal then
    Code := SB_HORZ
  else
    Code := SB_VERT;

  OldPos    := fPosition;
  fPosition := (Value div fIncrement)*fIncrement;
  if OldPos=fPosition then Exit;

  if fKind = sbHorizontal then
    fControl.ScrollBy(OldPos-Value,0)
  else
    fControl.ScrollBy(0,OldPos-Value);

  if GetScrollPos(fControl.Handle, Code) <> fPosition then
    SetScrollPos(fControl.Handle, Code, fPosition, True);

end;

procedure TCustomScrollBar.SetRange(Value: Integer);
begin
  if Value <> Range then begin
    fRange:=Value;
    if fRange < 0 then fRange := 0;
    fControl.UpdateScrollBars;
  end;
end;

procedure TCustomScrollBar.SetSize(Value: Integer);
const
  SysConsts: array[TScrollBarKind] of Integer = (SM_CYHSCROLL, SM_CYVSCROLL);
var
  NewValue: Integer;
begin
  if Value <> Size then begin
    NewValue := Value;
    if NewValue = 0 then Value := GetSystemMetrics(SysConsts[fKind]);
    fSize := Value;
    fControl.UpdateScrollBars;
    if NewValue = 0 then FSize := 0;
  end;
end;

procedure TCustomScrollBar.SetVisible(Value: boolean);
var
  Code: word;
begin
  fVisible:=Value;
  Code:=SB_HORZ;
  if fKind=sbVertical then Code:=SB_VERT;
  if fVisible and fControl.HandleAllocated then
    ShowScrollBar(fControl.Handle, Code, Visible)
  else
    fControl.UpdateScrollBars;
end;

procedure TCustomScrollBar.SetEnabled(Value: boolean);
var
  Code: word;
begin
  fEnabled:=Value;
  Code:=SB_HORZ;
  if fKind=sbVertical then Code:=SB_VERT;
  if fControl.HandleAllocated then
    if fEnabled then
      EnableScrollBar(fControl.Handle,Code,ESB_ENABLE_BOTH)
    else
      EnableScrollBar(fControl.Handle,Code,ESB_DISABLE_BOTH);
end;
{$EndRegion}

//######################################################################

constructor TWorkBitmapLayer.Create(OWner: TCustomImageView; Size: TPoint; ALayerCollection: TLayerCollection);
begin
  inherited Create(ALayerCollection);
  fControl      := OWner;

  fLeft         := (fControl.Width-Size.X*fControl.fScale) div 2;
  fTop          := (fControl.Height-Size.Y*fControl.fScale) div 2;

  if fControl.fScale>0 then begin
    fLeft:=(fLeft div fControl.fScale) * fControl.fScale;
    fTop:=(fTop div fControl.fScale) * fControl.fScale;
  end;

  fWidth        := Size.X;
  fHeight       := Size.Y;

  Location      := FloatRect(fLeft,fTop,fWidth*fControl.fScale+fLeft,fHeight*fControl.fScale+fTop);
  Scaled        := True;
  MouseEvents   := True;
  Tag           := fControl.fLayers.Count;
  Bitmap.Create;
  Bitmap.Clear;
  Bitmap.ResetAlpha(0);
  Bitmap.DrawMode := dmBlend;
  Bitmap.SetSize(Size.X,Size.Y);
  Bitmap.MasterAlpha:=$ff;
  Bitmap.FillRect(0,0,Size.X,Size.Y,$00ffffff);

end;

destructor TWorkBitmapLayer.Destroy;
begin
  inherited;
end;

procedure TWorkBitmapLayer.SetLeft(Value: integer);
begin
  if fLeft<>Value then begin
    fLeft:=Value;
    Location := FloatRect(fLeft,fTop,fLeft+fWidth*fControl.fScale,fTop+fHeight*fControl.fScale);
  end;
end;

procedure TWorkBitmapLayer.SetTop(Value: integer);
begin
  if fTop<>Value then begin
    fTop:=Value;
    Location := FloatRect(fLeft,fTop,fLeft+fWidth*fControl.fScale,fTop+fHeight*fControl.fScale);
  end;
end;

//######################################################################

constructor TCustomImageView.Create(AOwner: TComponent; Size: TPoint);
begin
  inherited Create(AOwner);
  fHorzScrollBar  := TCustomScrollBar.Create(TCustomImageView(Self), sbHorizontal);
  fVertScrollBar  := TCustomScrollBar.Create(TCustomImageView(Self), sbVertical);
  DoubleBuffered  := true;

  Width := 100;
  Height:= 100;

  fWidth := Size.X;
  fHeight:= Size.Y;

  fScale := 384 div fWidth;

  OnPaintStage:=PaintStage;
  if PaintStages[0]^.Stage = PST_CLEAR_BACKGND then PaintStages[0]^.Stage := PST_CUSTOM;

  fLayers:= TObjectList.Create;

  fTmpLayer              := TWorkBitmapLayer.Create(Self,Point(1,1),Layers);
  fTmpLayer.Location     := FloatRect(0,0,1,1);
  fTmpLayer.Scaled       := True;
  fTmpLayer.MouseEvents  := True;
  fTmpLayer.Tag          := 0;
  fTmpLayer.Bitmap.Create;
  fTmpLayer.Bitmap.Clear;
  fTmpLayer.Bitmap.ResetAlpha(0);
  fTmpLayer.Bitmap.DrawMode := dmBlend;
  fTmpLayer.Bitmap.SetSize(fWidth,fHeight);
  fTmpLayer.Bitmap.MasterAlpha:=$ff;
  fTmpLayer.Bitmap.FillRect(0,0,1,1,$00ffffff);
  fTmpLayer.Visible:=false;

  fActiveLayer := -1;
  fActiveColor := $ff000000;
  fDrawMode    := _select;

  fActiveCursor:= crDefault;
end;

destructor TCustomImageView.Destroy;
begin
  fHorzScrollBar.Free;
  fVertScrollBar.Free;
  fLayers.Free;
  inherited Destroy;  
end;

procedure TCustomImageView.SetHorzScrollBar(Value: TCustomScrollBar);
begin
  fHorzScrollBar.Assign(Value);
end;

procedure TCustomImageView.SetVertScrollBar(Value: TCustomScrollBar);
begin
  fVertScrollBar.Assign(Value);
end;

procedure TCustomImageView.SetActiveColor(Value: TColor32);
begin
  if Value<>fActiveColor then
    fActiveColor:=Value;
end;

procedure TCustomImageView.SetActiveLayer(Value: integer);
begin
  if Value<>fActiveLayer then begin
    fActiveLayer:=Value;
    invalidate;
  end;
end; 

procedure TCustomImageView.AddLayer(Width,Height: integer);
var
  fLayer : TWorkBitmapLayer;
begin
  fLayer       := TWorkBitmapLayer.Create(Self,Point(Width,Height),Layers);
  fLayers.Add(fLayer);
  fActiveLayer := fLayers.Count-1;
end;

function TCustomImageView.GetLayer(const LayerIndex: Integer): TWorkBitmapLayer;
begin
  Result := fLayers[LayerIndex] as TWorkBitmapLayer;
end;

procedure TCustomImageView.UpdateScrollBars;
begin
  if not HandleAllocated then Exit;
  if FVertScrollBar.NeedsScrollBarVisible then begin
    fHorzScrollBar.Update(False, True);
    fVertScrollBar.Update(True, False);
  end else if FHorzScrollBar.NeedsScrollBarVisible then begin
    fVertScrollBar.Update(False, True);
    fHorzScrollBar.Update(True, False);
  end else begin
    fVertScrollBar.Update(False, False);
    fHorzScrollBar.Update(True, False);
  end;
end;

procedure TCustomImageView.ScrollBy(DeltaX, DeltaY: Integer);
var
  R: TRect;
begin
  if (Handle=0) and not IsWindowVisible(Handle) then Exit;
  R:=Canvas.ClipRect;
  ScrollWindowEx(Canvas.Handle,DeltaX,DeltaY,nil,@r,0,nil,SW_ERASE or SW_INVALIDATE);
  Invalidate;
end;

procedure TCustomImageView.WMHScroll(var Message: TWMHScroll);
begin
  if (Message.ScrollBar = 0) and fHorzScrollBar.Visible then
    fHorzScrollBar.ScrollMessage(Message)
  else
    inherited;
end;

procedure TCustomImageView.WMVScroll(var Message: TWMVScroll);
begin
  if (Message.ScrollBar = 0) and fVertScrollBar.Visible then
    fVertScrollBar.ScrollMessage(Message)
  else
    inherited;
end;

procedure TCustomImageView.WMMouseWheel(var Message: TMessage);
var
  i: integer;
begin
  inherited;

  if Short(Message.WParamHi) > 0 then begin
    fScale:=fScale+1;
    fDirect:=1;
  end else begin
    fScale:=fScale-1;
    fDirect:=-1;
  end;

  if fScale<1 then begin fScale:=1; Exit; end;

  if fLayers.Count>0 then
    for i:=0 to fLayers.Count-1 do
      with TWorkBitmapLayer(fLayers[i]) do begin
        Left:= Left-CorrX*fDirect;
        Top := Top -CorrY*fDirect;
      end;

  fVertScrollBar.fRange:=fHeight*fScale;
  fHorzScrollBar.fRange:=fWidth*fScale;
  UpdateScrollBars;
  Invalidate;
end;

procedure TCustomImageView.WMSize(var Message: TWMSize);
begin
  inherited;
  fVertScrollBar.fRange:=fHeight*fScale;
  fHorzScrollBar.fRange:=fWidth*fScale;
  fVertScrollBar.Increment:=1;
  fHorzScrollBar.Increment:=1;
  if (GetWindowLong(Handle, GWL_STYLE) and WS_HSCROLL)=0 then HorzScrollBar.Position:=0;
  if (GetWindowLong(Handle, GWL_STYLE) and WS_VSCROLL)=0 then VertScrollBar.Position:=0;
  UpdateScrollBars;
  invalidate;
end;

procedure TCustomImageView.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS
                 or DLGC_WANTALLKEYS
                 or DLGC_WANTTAB;
end;

procedure TCustomImageView.WMPaint(var Message: TMessage);
const
  CrossColor : array [Boolean] of TColor = ($CCCCCC, $000000);
var
  DC: HDC;
  R : TRect;
  i : integer;
  Cl: TColor;

  function MakeRect(const FR: TFloatRect): TRect;
  begin
    Result.Left   := round(FR.Left);
    Result.Top    := round(FR.Top);
    Result.Right  := round(FR.Right);
    Result.Bottom := round(FR.Bottom);
  end;

begin
  inherited;
  if fActiveLayer<>-1 then begin
    DC     := GetDC(Self.Handle);
    R:=MakeRect(TWorkBitmapLayer(fLayers[fActiveLayer]).Location);

    for i:=1 to 11 do begin

      Cl:=CrossColor[i mod 2=0];

      SetPixel(DC,R.Left-1,R.Top-7+i,Cl);
      SetPixel(DC,R.Left-7+i,R.Top-1,Cl);

      SetPixel(DC,R.Left-1,R.Bottom-6+i,Cl);
      SetPixel(DC,R.Left-7+i,R.Bottom,Cl);

      SetPixel(DC,R.Right,R.Top-7+i,Cl);
      SetPixel(DC,R.Right-6+i,R.Top-1,Cl);

      SetPixel(DC,R.Right,R.Bottom-6+i,Cl);
      SetPixel(DC,R.Right-6+i,R.Bottom,Cl);
    end;

    ReleaseDC(0,DC);
  end;
end;

function TCustomImageView.MouseInRect(X,Y: integer): boolean;
begin
  Result:=false;
  if fActiveLayer<>-1 then
    with TWorkBitmapLayer(fLayers[fActiveLayer]) do
      Result:=(Location.Left<=X) and (Location.Right>=X) and (Location.Top<=Y) and (Location.Bottom>=Y);
end;

function TCustomImageView.MouseInRange(X,Y: integer): boolean;
begin
  with TWorkBitmapLayer(fLayers[fActiveLayer]) do
    Result:=(0<=X) and (fWidth-1>=X) and (0<=Y) and (fHeight-1>=Y);
end;

procedure TCustomImageView.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  MMx,MMy: integer;
begin
  MMx:=0;
  MMy:=0;
  
  if (fActiveLayer<>-1) then begin
    with TWorkBitmapLayer(fLayers[fActiveLayer]) do begin
      MMx:=(X-fLeft+HorzScrollBar.Position) div fScale;
      MMy:=(Y-fTop+VertScrollBar.Position) div fScale;
    end;
    if MouseInRange(MMx,MMy) then
      Cursor:=fActiveCursor
    else
      Cursor:=crDefault;
  end;

  if (fActiveLayer<>-1) and fDown then begin
    with TWorkBitmapLayer(fLayers[fActiveLayer]) do
      if fLock then begin
        Left:= ((X-Mx) div fScale) * fScale;
        Top := ((Y-My) div fScale) * fScale;
      end else begin
        EDx:=MMx;//(X-fLeft+HorzScrollBar.Position) div fScale;
        EDy:=MMy;//(Y-fTop+VertScrollBar.Position) div fScale;

        if MouseInRange(EDx,EDy) and MouseInRect(X,Y) then begin
          BeginUpdate;
          ImageDraw(fTmpLayer.Bitmap,SDx,SDy,EDx,EDy,NDx,NDy,fActiveColor,fDrawMode);
          EndUpdate;
        end;

        if EDx<0 then Ndx:=0 else NDx:=EDx;
        if EDx>=fWidth-1 then Ndx:=fWidth-1 else NDx:=EDx;
        if EDy<0 then Ndy:=0 else NDy:=EDy;
        if EDy>=fHeight-1 then Ndy:=fHeight-1 else NDy:=EDy;
      end;
    Invalidate;
  end;
end;

procedure TCustomImageView.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (fActiveLayer<>-1) and MouseInRect(X,Y) then begin
    fDown:=true;
    with TWorkBitmapLayer(fLayers[fActiveLayer]) do
      if fLock then begin
        Mx:=X-Left;
        My:=Y-Top;
      end else begin
        fTmpLayer.Visible:=true;
        fTmpLayer.Index:=Index+1;
        fTmpLayer.Location:=Location;
        fTmpLayer.Bitmap:=Bitmap;
        fTmpLayer.Bitmap.Clear($00ffffff);
        SDx:=(X-fLeft+HorzScrollBar.Position) div fScale;
        SDy:=(Y-fTop+VertScrollBar.Position) div fScale;
        fTmpLayer.Bitmap.PixelS[SDx,SDy]:=fActiveColor;
        NDx:=SDx;
        NDy:=SDy;
      end;
  end;
  Invalidate;
end;

procedure TCustomImageView.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (fActiveLayer<>-1) and fDown then
    with TWorkBitmapLayer(fLayers[fActiveLayer]) do
      if fLock and MouseInRect(X,Y) then begin
        Left:= ((X-Mx) div fScale) * fScale;
        Top := ((Y-My) div fScale) * fScale;
      end else begin
        fTmpLayer.Bitmap.CombineMode:=cmMerge;
        fTmpLayer.Bitmap.DrawTo(Bitmap);
        fTmpLayer.Bitmap.Clear;
        fTmpLayer.Visible:=false;
      end;
  fDown:=false;
end;

procedure TCustomImageView.SaveToFile(FName: string);
var
  i,x,y: integer;
begin

  fTmpLayer.Bitmap.Clear;
  fTmpLayer.Bitmap.SetSize(fWidth,fHeight);
  fTmpLayer.Bitmap.CombineMode:=cmMerge;
  fTmpLayer.Bitmap.Clear($00ffffff);

  for i:=0 to fLayers.Count-1 do
    with TWorkBitmapLayer(fLayers[i]) do begin
      x:=(Left-OffsetX) div fScale;
      y:=(Top-OffsetY)  div fScale;
      fTmpLayer.Bitmap.Draw(x,y,Bitmap);
    end;
      
  fTmpLayer.Bitmap.SaveToFile(FName);
  
  fTmpLayer.Visible:=false;
  fTmpLayer.Bitmap.Clear($00ffffff);
  
end;

procedure TCustomImageView.PaintStage(Sender: TObject; Buffer: TBitmap32; StageNum: Cardinal);
const
  Colors: array [Boolean] of TColor = ($FFFFFF, $CCCCCC);
var
  X,Y,xx,yy,i     : integer;
  r               : TRect;
  OddY            : Integer;
begin

  if HandleAllocated and (Parent<>NIL) then
    if not Focused then SetFocus;

  Buffer.FillRectS(Buffer.ClipRect,color32(Color));

  X:=HorzScrollBar.Position;
  Y:=VertScrollBar.Position;

  OffsetX:=(Width-fWidth*fScale) div 2;
  OffsetY:=(Height-fHeight*fScale) div 2;
  if fScale>0 then begin
    OffsetX:=(OffsetX div fScale) * fScale;
    OffsetY:=(OffsetY div fScale) * fScale;
  end;

  with Buffer do
    for yy:=0 to fHeight-1 do begin
      OddY := yy and $1;
      for xx:=0 to fWidth-1 do begin
        r.Left  := xx*fScale+OffsetX-X;
        r.Top   := yy*fScale+OffsetY-Y;
        r.Right := r.Left+fScale;
        r.Bottom:= r.Top +fScale;
        FillRectS(r,Colors[xx and $1=OddY]);
      end;
    end;

  if fLayers.Count>0 then
    for i:=0 to fLayers.Count-1 do
      with TWorkBitmapLayer(fLayers[i]) do begin
        Left:= Left+(OffsetX-Ox);
        Top := Top +(OffsetY-Oy);
        CorrX:=((OffsetX-Left) div fScale);
        CorrY:=(OffsetY-Top)  div fScale;
        Location := FloatRect(fLeft-X,fTop-Y,fLeft+fWidth*fScale-X,fTop+fHeight*fScale-Y);
        //Buffer.Textout(0,0,IntToStr(fWidth));
      end;

  Ox:=OffsetX;
  Oy:=OffsetY;

end;

end.
