unit SlideBar;

interface

{$R SlideBar.res}  {SlideIco.RES}

{$A+} { Word align data }
{$B-} { Complete Boolean Evaluation Directive }
{$D-} { Debug Information Directive }
{$L-} { Local Symbol Information Directive }
{$R-} { Range checking off }
{$S-} { Stack checking off }
{$T-} { We don't need (nor want) this type checking! }

uses
  {$IFDEF WIN32}
    Windows,
  {$ELSE}
    WinTypes, WinProcs,
  {$ENDIF}
  SysUtils, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, Menus;

type
  {------------------------------------------------------------------------}
  TBarStyle    = (bsLowered, bsRaised);
  TOrientation = (orVertical, orHorizontal);
  TThumbStyle  = (tsBar1, tsBar2, tsBar3, tsBar4, tsCircle1, tsSquare1,
								  tsDiamond1, tsDiamond2, tsDiamond3, tsDiamond4,
                   {AFS} tsStdDown, tsStdUp, tsStdVBar, tsStdUpDown,
                   	tsStdLeft, tsStdRight, tsStdHBar, stStdLeftRight);

  TSlideBar = class(TCustomControl)
  private

    FFocusColor             : TColor;
    FHandCursor             : Boolean;
    FLabels                 : TStringList;
    FMax,FMin,FPosition     : Integer;
    FOrientation            : TOrientation;
    FStyle                  : TBarStyle;
    FThickness              : Byte;
    FThumbStyle             : TThumbStyle;
    FTicks                  : Boolean;
    FTicksColor             : TColor;
    FFlashColor             : Boolean;

    { AFS }
    FFrequency              : integer;
    FStepLine               : integer;
    FStepPage               : integer;

    FOnChange               : TNotifyEvent;
    ThumbBmp,MaskBmp,BkgdBmp: TBitmap;
    DragVal,HalfTW,HalfTH   : Integer;
    ThumbRect               : TRect;
    HandPointer             : HCursor;
    procedure SetLabels(Value: TStringList);
    procedure SetMax(Value: Integer);
    procedure SetMin(Value: Integer);
    procedure SetOrientation(Value: TOrientation);
    procedure SetPosition(Value: Integer);
    procedure SetStyle(Value: TBarStyle);
    procedure SetThickness(Value: Byte);
    procedure SetThumbStyle(Value: TThumbStyle);
    procedure SetTicks(Value: Boolean);
    procedure SetFlashColor(Value: Boolean);

    { AFS }
    procedure SetFrequency (Value: integer);
    procedure SetStepLine (Value: integer);
    procedure SetStepPage (Value: integer);

    procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  protected
    Dragging: Boolean;
    procedure Paint; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    function  NewPosition(WhereX,WhereY: Integer): Integer;
    function  IsVert: Boolean;
    procedure RemoveThumbBar;
    procedure DrawThumbBar;
    procedure DrawTrench;
    procedure SaveBackground;
    procedure WhereIsBar;
    procedure SetTLColor;
    procedure SetBRColor;
    procedure SetTickColor(Value:TColor);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CurrentLabel: String;

    { AFS }
    procedure MoveDown (iDist: integer);
		procedure MoveUp (iDist: integer);

  published
    property Enabled;
    property FocusColor: TColor read FFocusColor write FFocusColor default clSilver;
    property TicksColor: TColor read FTicksColor write SetTickColor default clGray;
    property HandCursor: Boolean read FHandCursor write FHandCursor default True;
    property Labels: TStringList read FLabels write SetLabels;
    property Max: Integer read FMax write SetMax default 10;
    property Min: Integer read FMin write SetMin default 1;
    property Orientation: TOrientation read FOrientation write SetOrientation default orVertical;
    property ParentShowHint;
    property Position: Integer read FPosition write SetPosition default 1;
    property PopupMenu;
    property ShowHint;
    property Style: TBarStyle read FStyle write SetStyle default bsLowered;
    property TabStop default True;
    property TabOrder;
    property Thickness: Byte read FThickness write SetThickness default 1;
    property ThumbStyle: TThumbStyle read FThumbStyle write SetThumbStyle default tsBar3;
    property Ticks: Boolean read FTicks write SetTicks default True;
    property FlashColor: Boolean read FFlashColor write SetFlashColor default False;

    { AFS }
    property Frequency : integer read FFrequency write SetFrequency default 1;
    property StepLine	: integer read FStepLine write SetStepLine default 1;
    property StepPage	: integer read FStepPage write SetStepPage;

    property Visible;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
  end;

procedure Register;

implementation


function MinInt(A, B: Integer): Integer;
begin
  If A > B Then MinInt := B Else MinInt := A;
end;

function MaxInt(A,B: Integer): Integer;
begin
  If A > B Then MaxInt := A Else MaxInt := B;
end;

(*
 * TSlideBar Methods
 *)

constructor TSlideBar.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height := 100;
  Width := 15;
  ThumbBmp := TBitmap.Create;
  MaskBmp := TBitmap.Create;
  BkgdBmp := TBitmap.Create;
  HandPointer := LoadCursor(HInstance, 'HandPointer');
  FFocusColor := clSilver;
  FTicksColor := clGray;
  FHandCursor := False;
  FLabels := TStringList.Create;
  FMin := 0;
  FMax := 100;
  FOrientation := orVertical;
  FPosition := 1;
  FStyle := bsLowered;
  FThickness := 2;
  FTicks := True;
  FFlashColor := False;
  { AFS }
  FFrequency := 2;
  FStepLine := 1;
  FStepPage := 10;
  Dragging := False;
  DragVal := 0;
  ThumbStyle := tsBar3;
  TabStop := True;
end;

destructor TSlideBar.Destroy;
begin
  FLabels.Free;
  ThumbBmp.Free;
  MaskBmp.Free;
  BkgdBmp.Free;
  inherited Destroy;
end;

procedure TSlideBar.CMEnter(var Message: TCMGotFocus);
begin
  inherited;
  Refresh;
end;

procedure TSlideBar.CMExit(var Message: TCMExit);
begin
  inherited;
  Refresh;
end;

function TSlideBar.IsVert: Boolean;
begin
  IsVert := (Orientation = orVertical);
end;

procedure TSlideBar.KeyDown (var Key: Word; Shift: TShiftState);
begin

	{ AFS - Factored in StepPage and StepLine
   created MoveUp & MoveDown public procedures }
  case Key of
    VK_PRIOR : MoveUp (StepPage); {Effect that PGup/PGDown have}
    VK_NEXT  : MoveDown (StepPage);
    VK_END   : {if IsVert then Position := Min else} Position := Max;
    VK_HOME  : {if IsVert then Position := Max else} Position := Min;
    VK_LEFT  : MoveDown (StepLine); {Effect that Arrow Keys have}
    VK_DOWN  : MoveDown (StepLine);
    VK_UP    : MoveUp (StepLine);
    VK_RIGHT : MoveUp (StepLine);
  end; { Case }

end; { KeyDown }

procedure TSlideBar.MoveDown (iDist: integer);
begin
	if iDist = 0 then
  	exit;

	if iDist < 0 then
  begin
  	MoveUp (-iDist);
    exit;
  end;

  if (Position - iDist) > Min then
    Position := Position - iDist
  else
    Position := Min;
end;

procedure TSlideBar.MoveUp (iDist: integer);
begin
	if iDist = 0 then
  	exit;

	if iDist < 0 then
  begin
  	MoveDown (-iDist);
    exit;
  end;

  if (Position + iDist) < Max then
    Position := Position + iDist
  else
    Position := Max;
end;

procedure TSlideBar.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;
end;

procedure TSlideBar.WMSetCursor(var Message: TWMSetCursor);
  begin
  if (not (csDesigning in ComponentState)) and HandCursor then
    {$IFDEF WIN32}
    Windows.SetCursor(HandPointer)
	  {$ELSE}
    SetCursor(HandPointer)
    {$ENDIF}
  else
    inherited;
  end;

procedure TSlideBar.WMSize(var Message: TWMSize);
begin
  if Height > Width then
    Orientation := orVertical
  else
    Orientation := orHorizontal;
end;

procedure TSlideBar.SetLabels(Value: TStringList);
begin
  FLabels.Assign(Value);
end;

procedure TSlideBar.SetMin(Value: Integer);
begin
  FMin := Value;
  { AFS - set page size to 1/10 of the range }
  StepPage := MaxInt(1,(Max-Min) div 10);
  Refresh;
end;

procedure TSlideBar.SetMax(Value: Integer);
begin
  FMax := Value;            
  { AFS - set page size to 1/10 of the range }
  StepPage := MaxInt(1,(Max-Min) div 10);
  Refresh;
end;

procedure TSlideBar.SetOrientation(Value: TOrientation);
var StoreH, StoreW : Integer;
begin
  StoreH:=Height;
  StoreW:=Width;
     Case Value of
     orHorizontal: if StoreH > StoreW then
                   begin
                   Height:=StoreW;
                   Width:=StoreH;
                   {Change the bitmap with the orientation to maintain the aspect}
                     if ThumbStyle=tsBar4 then ThumbStyle:=tsBar2;
                     if ThumbStyle=tsBar3 then ThumbStyle:=tsBar1;
                     if ThumbStyle=tsStdHBar then ThumbStyle:=tsStdVBar;
                     if ThumbStyle=tsStdRight then ThumbStyle:=tsStdDown;
                   end;

     orVertical:   if StoreH < StoreW then
                   begin
                   Height:=StoreW;
                   Width:=StoreH;
                   {Change the bitmap with the orientation to maintain the aspect}
                     if ThumbStyle=tsBar2 then ThumbStyle:=tsBar4;
                     if ThumbStyle=tsBar1 then ThumbStyle:=tsBar3;
                     if ThumbStyle=tsStdVBar then ThumbStyle:=tsStdHBar;
                     if ThumbStyle=tsStdDown then ThumbStyle:=tsStdRight;
                   end;
     end;
  FOrientation := Value;
  Refresh;
end;

procedure TSlideBar.SetPosition(Value: Integer);
var
  R, G, B: Byte;
  TrackColor, NP: Integer; {NP=Normalized Position}
begin
NP:=Trunc((768/(FMax-FMin+1))*Position)+256; {Zero based Min/Max}

  R:=0;
  G:=0;
  B:=0;

  if csDesigning in ComponentState then
    begin

        if FFlashColor = True then
        begin
          case NP  of
            0..255: begin R:=0;G:=NP;B:=255; end;
            256..511: begin R:=0;G:=255;B:=511-NP; end;
            512..767: begin R:=NP-512;G:=255;B:=0; end;
            768..1023: begin R:=255;G:=1024-NP;B:=0; end;
            1024..1279: begin R:=255;G:=0;B:=NP-1024; end;
            1280..1536: begin R:=1536-NP;G:=0;B:=255; end;
          end;
           TrackColor := R or (G shl 8) or (B shl 16);
           FFocusColor:=TrackColor;
        end;

      if (Value >= Min) and (Value <= Max) Then FPosition := Value;
      Refresh;
    end
  else
    begin

        if FFlashColor = True then
        begin
          case NP  of
            0..255: begin R:=0;G:=NP;B:=255; end;
            256..511: begin R:=0;G:=255;B:=511-NP; end;
            512..767: begin R:=NP-512;G:=255;B:=0; end;
            768..1023: begin R:=255;G:=1024-NP;B:=0; end;
            1024..1279: begin R:=255;G:=0;B:=NP-1024; end;
            1280..1536: begin R:=1536-NP;G:=0;B:=255; end;
          end;
           TrackColor := R or (G shl 8) or (B shl 16);
           FFocusColor:=TrackColor;
           Refresh;
         end;

      RemoveThumbBar;
      if (Value >= Min) and (Value <= Max) Then FPosition := Value;
      WhereIsBar;
      SaveBackground;
      DrawThumbBar;
      if Assigned(FOnChange) then FOnChange(Self);
   end;
end;

procedure TSlideBar.SetStyle(Value: TBarStyle);
begin
  FStyle := Value;
  Refresh;
end;

procedure TSlideBar.SetThickness(Value: Byte);
begin
  If (Value > 0) and (Value <= 12) then
  begin
    FThickness := Value;
    Refresh;
  end;
end;

procedure TSlideBar.SetThumbStyle(Value: TThumbStyle);
begin
  If ThumbStyle <> Value then
    begin
      FThumbStyle := Value;
      case ThumbStyle of
        tsBar1     :   ThumbBmp.Handle := LoadBitmap(HInstance,'Bar1');
        tsBar2     :   ThumbBmp.Handle := LoadBitmap(HInstance,'Bar2');
        tsBar3     :   ThumbBmp.Handle := LoadBitmap(HInstance,'Bar3');
        tsBar4     :   ThumbBmp.Handle := LoadBitmap(HInstance,'Bar4');
        tsCircle1  :   ThumbBmp.Handle := LoadBitmap(HInstance,'Circle1');
        tsSquare1  :   ThumbBmp.Handle := LoadBitmap(HInstance,'Square1');
        tsDiamond1 :   ThumbBmp.Handle := LoadBitmap(HInstance,'Diamond1');
        tsDiamond2 :   ThumbBmp.Handle := LoadBitmap(HInstance,'Diamond2');
        tsDiamond3 :   ThumbBmp.Handle := LoadBitmap(HInstance,'Diamond3');
        tsDiamond4 :   ThumbBmp.Handle := LoadBitmap(HInstance,'Diamond4');
        { AFS }
        tsStdDown  :	 ThumbBmp.Handle := LoadBitmap(HInstance,'StdDown');
        tsStdUp  	 :	 ThumbBmp.Handle := LoadBitmap(HInstance,'StdUp');
        tsStdVBar  :	 ThumbBmp.Handle := LoadBitmap(HInstance,'StdVBar');
        tsStdUpDown:	 ThumbBmp.Handle := LoadBitmap(HInstance,'StdUpDown');
        tsStdLeft  :   ThumbBmp.Handle := LoadBitmap(HInstance,'StdLeft');
        tsStdRight :   ThumbBmp.Handle := LoadBitmap(HInstance,'StdRight');
        tsStdHBar  :   ThumbBmp.Handle := LoadBitmap(HInstance,'StdHBar');
        stStdLeftRight:ThumbBmp.Handle := LoadBitmap(HInstance,'StdLeftRight');
	else   Raise Exception.Create ('TSlideBar.SetThumbStyle: Bad value');

      end; { case }

      case ThumbStyle of
        tsBar1     :   MaskBmp.Handle := LoadBitmap(HInstance,'Bar1Mask');
        tsBar2     :   MaskBmp.Handle := LoadBitmap(HInstance,'Bar2Mask');
        tsBar3     :   MaskBmp.Handle := LoadBitmap(HInstance,'Bar3Mask');
        tsBar4     :   MaskBmp.Handle := LoadBitmap(HInstance,'Bar4Mask');
        tsCircle1  :   MaskBmp.Handle := LoadBitmap(HInstance,'Circle1Mask');
        tsSquare1  :   MaskBmp.Handle := LoadBitmap(HInstance,'Square1Mask');
        tsDiamond1 :   MaskBmp.Handle := LoadBitmap(HInstance,'Diamond1Mask');
        tsDiamond2 :   MaskBmp.Handle := LoadBitmap(HInstance,'Diamond2Mask');
        tsDiamond3 :   MaskBmp.Handle := LoadBitmap(HInstance,'Diamond3Mask');
        tsDiamond4 :   MaskBmp.Handle := LoadBitmap(HInstance,'Diamond4Mask');
        { AFS }
        tsStdDown  :	 MaskBmp.Handle := LoadBitmap(HInstance,'StdDownMask');
        tsStdUp    :	 MaskBmp.Handle := LoadBitmap(HInstance,'StdUpMask');
        tsStdVBar   :	 MaskBmp.Handle := LoadBitmap(HInstance,'StdVBarMask');
        tsStdUpDown:	 MaskBmp.Handle := LoadBitmap(HInstance,'StdUpDownMask');
	      tsStdLeft  :   MaskBmp.Handle := LoadBitmap(HInstance,'StdLeftMask');
        tsStdRight :   MaskBmp.Handle := LoadBitmap(HInstance,'StdRightMask');
        tsStdHBar  :   MaskBmp.Handle := LoadBitmap(HInstance,'StdHBarMask');
        stStdLeftRight:MaskBmp.Handle := LoadBitmap(HInstance,'StdLeftRightMask');
	else  Raise Exception.Create ('TSlideBar.SetThumbStyle: Bad value');
      end; { case }

      HalfTH := ThumbBmp.Height div 2;
      HalfTW := ThumbBmp.Width div 2;
      Refresh;
    end;
end;

procedure TSlideBar.SetTicks(Value: Boolean);
begin
	if FTicks = Value then
  	exit; { AFS}
  FTicks := Value;
  Refresh;
end;

procedure TSlideBar.SetFlashColor(Value: Boolean);
begin
if FFlashColor = Value then exit; {Don't do it unnecessarily}{MN}
  FFlashColor := Value;
  {Restore the default settings}
  If FFlashColor=False then FFocusColor:=clSilver else FFocusColor:=$00F8FF00;
  Refresh;
end;

{ AFS }
procedure TSlideBar.SetFrequency (Value: integer);
begin
	if FFrequency = Value then
  	exit;
  FFrequency := Value;
  Refresh;
end; {procedure SetFrequency }

procedure TSlideBar.SetStepLine (Value: integer);
begin
	if FStepLine = Value then
  	exit;
  if (Value < 1) or (Value > (Max - Min)) then
  	exit;

  FStepLine := Value;

  if StepPage < StepLine then
  	FStepPage := StepLine;
end; {procedure SetStepLine }
                                             
procedure TSlideBar.SetStepPage (Value: integer);
begin
	if FStepPage = Value then
  	exit;
  if (Value < 1) or (Value > (Max - Min)) then
  	exit;

  FStepPage := Value;

  if StepLine > StepPage then
  	FStepLine := StepPage;
end; { procedure SetStepPage }


function TSlideBar.CurrentLabel: String;
begin
  if ((Position-Min+1) <= Labels.Count) and (Position >= Min) then
    CurrentLabel := Labels[Position-Min]
  else
    CurrentLabel := '<Un-Defined>';
end;

function TSlideBar.NewPosition(WhereX,WhereY: Integer): Integer;
var
  H1,W1 : Integer;
begin
  {Calculate the nearest position to where the mouse is located}
  H1 := Height-HalfTH;
  W1 := Width-HalfTW;
  if IsVert then
    Result := Round(((H1-WhereY)/H1)*(Max-Min)+Min)
  else
    Result := Round((WhereX/W1)*(Max-Min)+Min);
  Result := MinInt(MaxInt(Result,Min),Max);
end;

procedure TSlideBar.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  A,B,C,D,E : Integer;
begin
  if Button <> mbLeft then exit;
  C := Position-1;
  D := Position;
  E := Position+1;
  {B is the center of the ThumbBar}
  if IsVert then B := ThumbRect.Top+HalfTH else B := ThumbRect.Left+HalfTW;
  if Dragging then
    A := NewPosition(X,Y)
  else
    if IsVert then
      if Y < B then A := E else if Y > B then A := C else A := D
    else
      if X < B then A := C else if X > B then A := E else A := D;
  A := MinInt(MaxInt(A,Min),Max);
  Dragging := False;
  Position := A;
end;

procedure TSlideBar.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  SetFocus;
  Dragging := PtInRect(ThumbRect,Point(X,Y));
  If IsVert then DragVal := Y else DragVal := X;
end;

procedure TSlideBar.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  LastDragVal : Integer;
begin
  {Is the left mouse button down and dragging the thumb bar?}
  if (ssLeft in Shift) and Dragging then
    begin
      LastDragVal := DragVal;
      if IsVert then DragVal := Y else DragVal := X;
      {This test eliminates unneccesary repaints}
      if DragVal <> LastDragVal then Position := NewPosition(X,Y);
    end;
end;

procedure TSlideBar.RemoveThumbBar;
begin
  {Place the background bitmap where it was last}
  Canvas.Draw(ThumbRect.Left,ThumbRect.Top,BkgdBmp);
end;

procedure TSlideBar.DrawThumbBar;
var
  TmpBmp  : TBitMap;
  Rect1   : TRect;
begin
  { *> moved here to avoid a warning message }
  {   AFS: this is correct -
     	if there is an exception in the objects's constructor
      this you do not want to try to free the object}

  {Create a working bitmap}
  TmpBmp := TBitmap.Create;
  try
    {Define a rectangle to mark the dimensions of the thumbbar}
    Rect1 := Rect (0,0,ThumbBmp.Width,ThumbBmp.Height);
    TmpBmp.Height := ThumbBmp.Height;
    TmpBmp.Width := ThumbBmp.Width;
    {Copy the background area onto the working bitmap}
    TmpBmp.Canvas.CopyMode := cmSrcCopy;
    TmpBmp.Canvas.CopyRect(Rect1,BkgdBmp.Canvas,Rect1);
    {Copy the mask onto the working bitmap with SRCAND}
    TmpBmp.Canvas.CopyMode := cmSrcAnd;
    TmpBmp.Canvas.CopyRect(Rect1,MaskBmp.Canvas,Rect1);
    {Copy the thumbbar onto the working bitmap with SRCPAINT}
    TmpBmp.Canvas.CopyMode := cmSrcPaint;
    TmpBmp.Canvas.CopyRect(Rect1,ThumbBmp.Canvas,Rect1);
    {Now draw the thumb bar}
    Canvas.CopyRect(ThumbRect,TmpBmp.Canvas,Rect1);
  finally
    TmpBmp.Free;
  end;
end;

{ Made vertical like Win95 - zero at the top <- that's silly => moved the zero at the bottom like in any normal mixer!!! (MN)}
procedure TSlideBar.WhereIsBar;
var
  Each          : Single;
  ThumbX,ThumbY : Integer;
begin
  {Calculate where to paint the thumb bar - stored in ThumbRect}
  if IsVert then
    begin
      Each := (Height-ThumbBmp.Height)/(Max-Min);
      If Dragging then
				ThumbY := DragVal-HalfTH
      else
				ThumbY := Height-Round(Each*(Position-Min))-ThumbBmp.Height;
      ThumbY := MaxInt(0,MinInt(Height-ThumbBmp.Height,ThumbY));
      ThumbX := (Width-ThumbBmp.Width) div 2;
    end
  else
    begin
      Each := (Width-ThumbBmp.Width)/(Max-Min);
      if Dragging then
				ThumbX := DragVal-HalfTW
      else
      	ThumbX := Round(Each*(Position-Min));
      ThumbX := MaxInt(0,MinInt(Width-ThumbBmp.Width,ThumbX));
      ThumbY := (Height-ThumbBmp.Height) div 2;
    end;

  ThumbRect := Rect(ThumbX,ThumbY,ThumbX+ThumbBmp.Width,ThumbY+ThumbBmp.Height);

end; { procedure WhereIsThumb }

procedure TSlideBar.SetTLColor;
begin
  {Set the Top/Left color for the trench. Controls raised or lowered styles clGray/clWhite}
  With Canvas do
    if Style = bsLowered then Pen.Color := $2C2C2C else Pen.Color := clWhite;
end;

procedure TSlideBar.SetBRColor;
begin
  {Set the Bottom/Right color for the trench. Controls raised or lowered styles clGray/clWhite}
  With Canvas do
    if Style = bsRaised then Pen.Color := $2C2C2C else Pen.Color := clWhite;
end;

procedure TSlideBar.SetTickColor(Value:TColor);
begin
	if FTicksColor = Value then
  	exit; {MN}
        FTicksColor := Value;
    Canvas.Pen.Color := FTicksColor;
    Refresh;
end;

{ AFS - fixed intendation and did tick frequency }
procedure TSlideBar.DrawTrench;
var
  X1,Y1,X2,Y2 : Integer;
  Each        : Single;
  iTickLoop,TickPos : Integer;
begin
  {This procedure simply draws the slot that the thumb bar will travel through}
  {including the tick-marks. The bar itself is not drawn.}
  with Canvas do
  begin
    {Calculate the corners of the trench dependant on orientation}
    if IsVert then
    begin
      X1 := (Width div 2) - (Thickness div 2) - 1;
      X2 := X1 + Thickness + 1;
      Y1 := HalfTH;
      Y2 := Height-ThumbBmp.Height+Y1;
    end { IsVert }
    else
    begin
      X1 := HalfTW;
      X2 := Width-ThumbBmp.Width+X1;
      Y1 := (Height div 2) - (Thickness div 2) - 1;
      Y2 := Y1 + Thickness + 1;
    end; { not vert }

    Pen.Style := psSolid;

    SetTLColor;
    MoveTo(X2,Y1);
    LineTo(X1,Y1);
    LineTo(X1,Y2);

    SetBRColor;
    LineTo(X2,Y2);
    LineTo(X2,Y1-1);

    with brush do {if Focused then} Color := FocusColor {else Color := clSilver};
    Pen.Style := psClear;
    Rectangle(X1+1,Y1+1,X2+1,Y2+1);
    Pen.Style := psSolid;
    Each := 0;

    if Ticks then begin
    
      if (Max-Min) > 0 then
				if IsVert then
	  			Each := (Height-ThumbBmp.Height)/(Max-Min)
				else
	  			Each := (Width-ThumbBmp.Width)/(Max-Min);

      for iTickLoop := Min to Max do begin
      	if (iTickLoop mod Frequency) <> 0 then Continue;
        if IsVert then begin
            TickPos := Y2-Trunc(Each*(iTickLoop - Min)) -1;
            if iTickLoop = Max then TickPos := Y1;
            Canvas.Pen.Color := FTicksColor;
            MoveTo(X1-3,TickPos);            {draw them on one side}
            LineTo(X1-7,TickPos);
            MoveTo(FThickness+X1+3,TickPos); {now draw them on the other side as well}
            LineTo(FThickness+X1+7,TickPos);
        end else begin
            TickPos := X1+Trunc(Each*(iTickLoop-Min));
            if iTickLoop = Max then TickPos := X2-1;
            Canvas.Pen.Color := FTicksColor;
            MoveTo(TickPos,Y1-3);            {draw them on one side}
            LineTo(TickPos,Y1-7);
            MoveTo(TickPos,FThickness+Y1+3); {now draw them on the other side as well}
            LineTo(TickPos,FThickness+Y1+7);
        end; { Not Vert }
    	end; { for loop }

		end; { Has ticks }

	end; { with canvas }

end; { DrawTrench }

procedure TSlideBar.SaveBackground;
begin
  BkgdBmp.Width := ThumbBmp.Width;
  BkgdBmp.Height := ThumbBmp.Height;
  BkgdBmp.Canvas.CopyRect(Rect(0,0,ThumbBmp.Width,ThumbBmp.Height), Canvas,ThumbRect);
end;

procedure TSlideBar.Paint;
begin
  DrawTrench;
  WhereIsBar;
  SaveBackground;
  DrawThumbBar;
end;

{------------------------------------------------------------------------}
procedure Register;
begin
  RegisterComponents('Samples', [TSlideBar]);
end;
{------------------------------------------------------------------------}

end.
